"""Converts a Neuroglancer state to display a cross-section cube."""

import argparse

import neuroglancer
import neuroglancer.cli
import neuroglancer.coordinate_space

if __name__ == "__main__":
    ap = argparse.ArgumentParser()
    ap.add_argument(
        "--cube-size",
        type=neuroglancer.coordinate_space.parse_unit_and_scale,
        default=(4e-6, "m"),
    )
    neuroglancer.cli.add_server_arguments(ap)
    neuroglancer.cli.add_state_arguments(ap, required=True)
    args = ap.parse_args()
    neuroglancer.cli.handle_server_arguments(args)

    state = args.state

    cube_size = args.cube_size[0]

    canonical_scale = min(d.scale for d in state.dimensions)

    state.show_axis_lines = False
    state.show_slices = False
    state.show_default_annotations = False
    state.projection_scale = 2 * cube_size / canonical_scale

    #  x+y+:  [0, 0, 0, 1]
    # z-y+:    [ 0,0.7071067690849304,0,0.7071067690849304]
    # z+y+:    [ 0,-0.7071067690849304,0,0.7071067690849304]
    # z+x-:   [0.5, -0.5, 0.5, 0.5]
    # z+x+:   [0.5, 0.5, 0.5, -0.5]
    # x+z-:   [ -0.7071067690849304,0, 0, 0.7071067690849304]

    orientations = [
        [0, 0.7071067690849304, 0, 0.7071067690849304],  # z-y+
        [0.5, 0.5, 0.5, -0.5],  # z+x+
        [0, 0, 0, 1],  # x+y+
    ]

    # Add 6 cube faces
    for face_dim in range(3):
        for face_dir in range(2):
            state.layout.type = "3d"
            position = list(state.position)
            position[face_dim] += (
                (face_dir * 2 - 1) * cube_size / 2 / state.dimensions[face_dim].scale
            )
            state.layout.cross_sections["%d_%d" % (face_dim, face_dir)] = (
                neuroglancer.CrossSection(
                    width=cube_size / canonical_scale,
                    height=cube_size / canonical_scale,
                    position=neuroglancer.LinkedPosition(
                        link="relative", value=position
                    ),
                    orientation=neuroglancer.LinkedOrientationState(
                        link="unlinked",
                        value=orientations[face_dim],
                    ),
                    scale=neuroglancer.LinkedZoomFactor(link="unlinked", value=1),
                )
            )

    print(neuroglancer.to_url(state))



# stdlib
import itertools

# externals
import numpy as np

# internals
from .utils import ensure_list


def volume_axis(*args, **kwargs):
    """Describe an axis of a volume of voxels.

    Signature
    ---------
    * ``def volume_axis(index, flipped=False)``
    * ``def volume_axis(name)``
    * ``def volume_axis(axis)``

    Parameters
    ----------
    index : () array_like[int]
        Index of the axis in 'direct' space (RAS)

    flipped : () array_like[bool], default=False
        Whether the axis is flipped or not.

    name : {'R' or 'L', 'A' or 'P', 'S' or 'I'}
        Name of the axis, according to the neuroimaging convention:
        * 'R' for *left to Right* (index=0, flipped=False) or
          'L' for *right to Left* (index=0, flipped=True)
        * 'A' for *posterior to Anterior* (index=1, flipped=False) or
          'P' for *anterior to Posterior* (index=1, flipped=True)
        * 'S' for *inferior to Superior* (index=2, flipped=False) or
          'I' for *superior to Inferior* (index=2, flipped=True)

    axis : (2,) array_like[int]
        ax[0] = index
        ax[1] = flipped
        Description of the axis.

    Returns
    -------
    axis : (2,) tensor[int]
        Description of the axis.
        * `ax[0] = index`
        * `ax[1] = flipped`

    """
    def axis_from_name(name):
        name = name.upper()
        if name == 'R':
            return np.asarray([0, 0], dtype="int")
        elif name == 'L':
            return np.asarray([0, 1], dtype="int")
        elif name == 'A':
            return np.asarray([1, 0], dtype="int")
        elif name == 'P':
            return np.asarray([1, 1], dtype="int")
        elif name == 'S':
            return np.asarray([2, 0], dtype="int")
        elif name == 'I':
            return np.asarray([2, 1], dtype="int")

    def axis_from_index(index, flipped=False):
        index = np.asarray(index).reshape([])
        flipped = np.asarray(flipped).reshape([])
        return np.asarray([index, flipped], dtype="int")

    def axis_from_axis(ax):
        ax = np.asarray(ax, dtype="int").flatten()
        if ax.size != 2:
            raise ValueError('An axis should have two elements. Got {}.'
                             .format(ax.size))
        return ax

    # Dispatch based on input types or keyword arguments
    args = list(args)
    if len(args) > 0:
        if isinstance(args[0], str):
            return axis_from_name(*args, **kwargs)
        else:
            args[0] = np.asarray(args[0])
            if args[0].numel() == 1:
                return axis_from_index(*args, **kwargs)
            else:
                return axis_from_axis(*args, **kwargs)
    else:
        if 'name' in kwargs.keys():
            return axis_from_name(*args, **kwargs)
        elif 'index' in kwargs.keys():
            return axis_from_index(*args, **kwargs)
        else:
            return axis_from_axis(*args, **kwargs)


# Mapping from (index, flipped) to axis name
_axis_names = [['R', 'L'], ['A', 'P'], ['S', 'I']]


def volume_axis_to_name(axis):
    """Return the (neuroimaging) name of an axis. Its index must be < 3.

    Parameters
    ----------
    axis : (2,) array_like

    Returns
    -------
    name : str

    """
    index, flipped = axis
    if index >= 3:
        raise ValueError('Index names are only defined up to dimension 3. '
                         'Got {}.'.format(index))
    return _axis_names[index][flipped]


def volume_layout(*args, **kwargs):
    """Describe the layout of a volume of voxels.

    A layout is characterized by a list of axes. See `volume_axis`.

    Signature
    ---------
    volume_layout(ndim=3)
    volume_layout(name)
    volume_layout(axes)
    volume_layout(index, flipped=False)

    Parameters
    ----------
    ndim : int, default=3
        Dimension of the space.
        This version of the function always returns a directed layout
        (identity permutation, no flips), which is equivalent to 'RAS'
        but in any dimension.

    name : str
        Permutation of axis names, according to the neuroimaging convention:
        * 'R' for *left to Right* (index=0, flipped=False) or
          'L' for *right to Left* (index=0, flipped=True)
        * 'A' for *posterior to Anterior* (index=1, flipped=False) or
          'P' for *anterior to Posterior* (index=1, flipped=True)
        * 'S' for *inferior to Superior* (index=2, flipped=False) or
          'I' for *superior to Inferior* (index=2, flipped=True)
        The number of letters defines the dimension of the matrix
        (`ndim = len(name)`).

    axes : (ndim, 2) array_like[int]
        List of objects returned by `axis`.

    index : (ndim, ) array_like[int]
        Index of the axes in 'direct' space (RAS)

    flipped : (ndim, ) array_like[bool], default=False
        Whether each axis is flipped or not.

    Returns
    -------
    layout : (ndim, 2) tensor[int]
        Description of the layout.

    """
    def layout_from_dim(dim):
        return volume_layout(list(range(dim)), flipped=False)

    def layout_from_name(name):
        return volume_layout([volume_axis(a) for a in name])

    def layout_from_index(index, flipped=False):
        index = np.asarray(index, dtype="int").flatten()
        ndim = index.shape[0]
        flipped = np.asarray(flipped, dtype="int").flatten()
        if flipped.shape[0] == 1:
            flipped = np.repeat(flipped, ndim, axis=0)
        return np.stack((index, flipped), axis=-1)

    def layout_from_axes(axes):
        axes = np.asarray(axes, dtype="int")
        if axes.ndim != 2 or axes.shape[1] != 2:
            raise ValueError('A layout should have shape (ndim, 2). Got {}.'
                             .format(axes.shape))
        return axes

    # Dispatch based on input types or keyword arguments
    args = list(args)
    if len(args) > 0:
        if isinstance(args[0], str):
            layout = layout_from_name(*args, **kwargs)
        else:
            args[0] = np.asarray(args[0])
            if args[0].ndim == 0:
                layout = layout_from_dim(*args, **kwargs)
            elif args[0].ndim == 2:
                layout = layout_from_axes(*args, **kwargs)
            else:
                layout = layout_from_index(*args, **kwargs)
    else:
        if 'dim' or 'ndim' in kwargs.keys():
            layout = layout_from_dim(*args, **kwargs)
        elif 'name' in kwargs.keys():
            layout = layout_from_name(*args, **kwargs)
        elif 'index' in kwargs.keys():
            layout = layout_from_index(*args, **kwargs)
        else:
            layout = layout_from_axes(*args, **kwargs)

    # Remap axes indices if not contiguous
    backend = dict(dtype=layout.dtype)
    axes = layout[:, 0]
    remap = np.argsort(axes)
    axes[remap] = np.arange(len(axes), **backend)
    layout = np.stack((axes, layout[:, 1]), axis=-1)
    return layout


def volume_layout_to_name(layout):
    """Return the (neuroimaging) name of a layout.

    Its length must be <= 3 (else, we just return the permutation and
    flips), e.g. '[2, 3, -1, 4]'

    Parameters
    ----------
    layout : (dim, 2) array_like

    Returns
    -------
    name : str

    """
    layout = volume_layout(layout)
    if len(layout) > 3:
        layout = [('-' if bool(f) else '') + str(int(p)) for p, f in layout]
        return '[' + ', '.join(layout) + ']'
    names = [volume_axis_to_name(axis) for axis in layout]
    return ''.join(names)


def iter_layouts(ndim):
    """Compute all possible layouts for a given dimensionality.

    Parameters
    ----------
    ndim : () array_like
        Dimensionality (rank) of the space.

    Returns
    -------
    layouts : (nflip*nperm, ndim, 2) tensor[int]
        All possible layouts.
        * nflip = 2 ** ndim     -> number of flips
        * nperm = ndim!         -> number of permutations

    """

    # First, compute all possible directed layouts on one hand,
    # and all possible flips on the other hand.
    axes = np.arange(ndim, dtype="int")
    layouts = np.asarray(itertools.permutations(axes))          # [P, D]
    flips = np.asarray(itertools.product([0, 1], repeat=ndim))  # [F, D]

    # Now, compute combination (= cartesian product) of both
    # We replicate each tensor so that shapes match and stack them.
    nb_layouts = len(layouts)
    nb_flips = len(flips)
    layouts = layouts[None, ...]
    layouts = np.repeat(layouts, nb_flips, axis=0)  # [F, P, D]
    flips = flips[:, None, :]
    flips = np.repeat(flips, nb_layouts, axis=1)    # [F, P, D]
    layouts = np.stack([layouts, flips], axis=-1)

    # Finally, flatten across repeats
    layouts = layouts.reshape([-1, ndim, 2])    # [F*P, D, 2]

    return layouts


def layout_matrix(layout, voxel_size=1., shape=0., dtype=None):
    """Compute the origin affine matrix for different voxel layouts.

    Resources
    ---------
    .. https://nipy.org/nibabel/image_orientation.html
    .. https://nipy.org/nibabel/neuro_radio_conventions.html

    Parameters
    ----------
    layout : str or (ndim, 2) array_like[int]
        See `affine.layout`

    voxel_size : (ndim,) array_like, default=1
        Voxel size of the lattice

    shape : (ndim,) array_like, default=0
        Shape of the lattice

    dtype : torch.dtype, optional
        Data type of the matrix

    Returns
    -------
    mat : (ndim+1, ndim+1) tensor[dtype]
        Corresponding affine matrix.

    """

    # Author
    # ------
    # .. Yael Balbastre <yael.balbastre@gmail.com>

    # Extract info from layout
    layout = volume_layout(layout)
    dim = len(layout)
    perm = invert_permutation(layout[:, 0])
    flip = layout[:, 1].astype("bool")

    # ensure tensor
    voxel_size = np.asarray(voxel_size)
    shape = np.asarray(shape)

    # ensure dim
    shape = np.asarray(ensure_list(shape.tolist(), dim))
    voxel_size = np.asarray(ensure_list(voxel_size.tolist(), dim))
    zero = np.zeros(dim, dtype=dtype)

    # Create matrix
    mat = np.diag(voxel_size)                               # [D, D]
    mat = mat[perm, :]                                      # [D, D]

    mflip = np.ones(dim)                                    # [D]
    mflip = np.where(flip, -mflip, mflip)                   # [D]
    mflip = np.diag(mflip)                                  # [D, D]
    shift = np.where(flip, shape[perm], zero)               # [D]
    mflip = np.concatenate((mflip, shift[:, None]), axis=1)         # [D, D+1]

    mat = np.matmul(mat, mflip)                             # [D, D+1]
    mat = np.concatenate([mat, [[0] * dim + [1]]], axis=0)            # [D+1, D+1]
    return mat


def invert_permutation(perm):
    """Return the inverse of a permutation

    Parameters
    ----------
    perm : (..., N) array_like
        Permutations. A permutation is a shuffled set of indices.

    Returns
    -------
    iperm : (..., N) array
        Inverse permutation.

    """
    perm = np.asarray(perm)
    shape = perm.shape
    perm = perm.reshape([-1, shape[-1]])
    n = perm.shape[-1]
    k = perm.shape[0]
    identity = np.arange(n, dtype="int")[None, ...]
    identity = np.broadcast_to(identity, [k, n])  # Repeat without allocation
    iperm = np.empty_like(perm)
    np.put_along_axis(iperm, perm, identity, -1)
    iperm = iperm.reshape(shape)
    return iperm



"""Tests for the `transforms` module."""
from ngtools import transforms as T
from ngtools.optionals import try_import_as

ng = try_import_as('neuroglancer', 'ngtools.nglite')


def test_convert_transform() -> None:
    """Test transform conversion."""
    # NOTE: we use `5 * 1E-4` and not `5E-4` because the conversion
    # code uses the former and both values differ in double precision
    # arithmetic.

    inp = ng.CoordinateSpaceTransform(
        input_dimensions=ng.CoordinateSpace({
            "i": [5, "um"],
            "j": [5, "mm"],
            "k": [5, "cm"],
            "t": [4, "ms"],
            "c": [1, ""],
        }),
        output_dimensions=ng.CoordinateSpace({
            "x": [1, "um"],
            "y": [2, "mm"],
            "z": [4, "cm"],
            "u": [6, "ms"],
            "c": [1, ""],
        }),
        matrix=[
            [1, 0, 0, 0, 0, 1],
            [0, 1, 0, 0, 0, 1],
            [0, 0, 1, 0, 0, 1],
            [0, 0, 0, 1, 0, 1],
            [0, 0, 0, 0, 1, 1],
        ]
    )

    ref = ng.CoordinateSpaceTransform(
        input_dimensions=ng.CoordinateSpace({
            "i": [5 * 10**-3,   "mm"],
            "j": [5,            "mm"],
            "k": [5 * 10,       "mm"],
            "t": [4 * 10**-3,   "s"],
            "c": [1,            ""],
        }),
        output_dimensions=ng.CoordinateSpace({
            "x": [1, "um"],
            "y": [2, "mm"],
            "z": [4, "cm"],
            "u": [6, "ms"],
            "c": [1, ""],
        }),
        matrix=[
            [1, 0, 0, 0, 0, 1],
            [0, 1, 0, 0, 0, 1],
            [0, 0, 1, 0, 0, 1],
            [0, 0, 0, 1, 0, 1],
            [0, 0, 0, 0, 1, 1],
        ]
    )
    out = T.convert_transform(inp, ["mm", "s"], None)
    assert out == ref, f"Convert to SI units:\nout={out}\nref={ref}"

    ref = ng.CoordinateSpaceTransform(
        input_dimensions=ng.CoordinateSpace({
            "i": [1, "m"],
            "j": [1, "m"],
            "k": [1, "m"],
            "t": [1, "s"],
            "c": [1, ""],
        }),
        output_dimensions=ng.CoordinateSpace({
            "x": [1, "m"],
            "y": [1, "m"],
            "z": [1, "m"],
            "u": [1, "s"],
            "c": [1, ""],
        }),
        matrix=[
            [5 * 1E-6, 0,        0,        0,        0, 1 * 1E-6],
            [0,        5 * 1E-3, 0,        0,        0, 2 * 1E-3],
            [0,        0,        5 * 1E-2, 0,        0, 4 * 1E-2],
            [0,        0,        0,        4 * 1E-3, 0, 6 * 1E-3],
            [0,        0,        0,        0,        1, 1],
        ]
    )
    out = T.normalize_transform(inp, unit_scale=True)
    assert out == ref, \
           f"Normalize transform (unit scale):\nout={out}\nref={ref}"


def test_inverse_compose() -> None:
    """Invert a transform and compose to recover identity."""
    inp = ng.CoordinateSpaceTransform(
        input_dimensions=ng.CoordinateSpace({
            "i": [5, "um"],
            "j": [5, "mm"],
            "k": [5, "cm"],
            "t": [4, "ms"],
            "c": [1, ""],
        }),
        output_dimensions=ng.CoordinateSpace({
            "x": [1, "um"],
            "y": [2, "mm"],
            "z": [4, "cm"],
            "u": [6, "ms"],
            "c": [1, ""],
        }),
        matrix=[
            [+0, -1, +0, +0, +0, +1],
            [+0, +0, +1, +0, +0, +1],
            [+1, +0, +0, +0, +0, +1],
            [+0, +0, +0, +1, +0, +1],
            [+0, +0, +0, +0, +1, +1],
        ]
    )

    inv = T.inverse(inp)

    ref_left = ng.CoordinateSpaceTransform(
        input_dimensions=inp.input_dimensions,
        output_dimensions=inp.input_dimensions,
        matrix=[
            [1, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0],
            [0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 1, 0],
        ]
    )
    out_left = T.compose(inv, inp)
    assert out_left == ref_left, \
           "Left-compose inverse:\nout={out_left}\nref={ref_left}"

    ref_right = ng.CoordinateSpaceTransform(
        input_dimensions=inp.output_dimensions,
        output_dimensions=inp.output_dimensions,
        matrix=[
            [1, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0],
            [0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 1, 0],
        ]
    )
    out_right = T.compose(inp, inv)
    assert out_right == ref_right, \
           "Right-compose inverse:\nout={out_right}\nref={ref_right}"



"""
Utilities to manipulate transforms (`ng.CoordinateSpaceTransform`) in
neuroglancer.

Also provide tools to read and convert arbitrary coordinate transforms into
neuroglancer format.

Functions
---------
compose
    Compose two `ng.CoordinateSpaceTransform`.
convert_transform
    Convert the units of a transform, while keeping the transform equivalent.
ensure_same_scale
    Convert the units of a transform, so that all axes have the same
    unit and scale, while keeping the transform equivalent.
get_matrix
    Get matrix from transform, creating one if needed.
inverse
    Invert a `ng.CoordinateSpaceTransform`.
load_affine
    Load an affine transform from a file or stream.
normalize_transform
    Ensure that input and output spaces have SI units.
subtransform
    Keep only axes whose units are of a given kind.
to_square
    Ensure that an affine matrix is square (with its homogeneous row).


Note on tranforms
-----------------

A `CoordinateSpaceTransform` has the following fields

    transform : np.ndarray
        Homogeneous affine transform that maps input coordinates
        (columns) to output coordinates (row)
    inputDimensions : CoordinateSpace
        A description of the input space.
        Sometimes referred to as the "source" space in the ng sourcecode.
    outputDimensions : CoordinateSpace
        A description of the output space.
        Sometimes referred to as the "view" space in the ng sourcecode.

A `CoordinateSpace` has the following fields:

    names : list[str]
        The name of each dimension.
    scales : list[float]
        The scale (voxel size) of each dimension.
    units : list[str]
        The unit in which scale is expressed.

Two important points must be noted, as they differ from how
transforms are encoded in other frameworks:

    * The _linear part_ of the transform is ___always___ unit-less.
        That is, it maps from/to the same isotropic unit.

    * The _translation_ component of the transform is ___always___
        expressed in output units.

A nifti-like "voxel to scaled-world" matrix would therefore be
constructed via

```
    [ x ]   [ 1/vx     ]   [ Axi Axj Axk ]   [ sx       ]   [ i ]   [ Tx ]
    [ y ] = [   1/vy   ] * [ Ayi Ayj Ayk ] * [    sy    ] * [ j ] + [ Ty ]
    [ z ]   [     1/vz ]   [ Azi Azj Azk ]   [       sz ]   [ k ]   [ Tz ]
```

where `v` is the view/output voxel size and `s` is the source/input
voxel size. In that context, [x, y, z] is expressed in "scaled world" units
and [i, j, k] is expressed in "voxel" units.

Alternatively, a "scaled-voxel to world" matrix can be constructed via

```
    [ x ]   [ Axi Axj Axk ]   [ sx       ]   [ i ]   [ vx       ]   [ Tx ]
    [ y ] = [ Ayi Ayj Ayk ] * [    sy    ] * [ j ] + [    vy    ] * [ Ty ]
    [ z ]   [ Azi Azj Azk ]   [       sz ]   [ k ]   [       vz ]   [ Tz ]
```
"""

# stdlib
import logging
import math
from numbers import Number
from os import PathLike
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import IO

# externals
import numpy as np

# internals
import ngtools.spaces as S
import ngtools.units as U
from ngtools._lta.lta import LinearTransformArray
from ngtools.opener import open, parse_protocols, stringify_path
from ngtools.optionals import try_from_import

# optionals
try:
    import neuroglancer as ng
except ImportError:
    import ngtools._nglite as ng
afni, fsl, itk, lta = try_from_import(
    'nitransforms.io', ['afni', 'fsl', 'itk', 'lta'],
    fallback="ngtools._ntlite.io",
)

LOG = logging.getLogger(__name__)

AFFINE_FORMATMAP = {}
if afni:
    AFFINE_FORMATMAP['afni'] = afni.AFNILinearTransform
if afni:
    AFFINE_FORMATMAP['fsl'] = fsl.FSLLinearTransform
if afni:
    AFFINE_FORMATMAP['itk'] = itk.ITKLinearTransform
if afni:
    if hasattr(lta, 'LinearTransformArray'):
        AFFINE_FORMATMAP['lta'] = lta.LinearTransformArray
    else:
        AFFINE_FORMATMAP['lta'] = lta.FSLinearTransform


def matrix_to_quaternion(mat: np.ndarray) -> np.ndarray:
    """Convert a rotation matrix to a quaternion.

    We follow neuroglancer and assume that quaternions are ordered as
    [*v, r] (or again, [i, j, k, r]). This differs form wikipedia
    which defines them as [r, *v] (or [r, i, j, k]).
    """
    # https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Fitting_quaternions
    mat = np.asarray(mat)
    if tuple(mat.shape) != (3, 3):
        raise ValueError("Expected a 3x3 matrix")
    if np.linalg.det(mat) < 0:
        # left-handed: flip first axis
        mat = np.copy(mat)
        mat[0, :] *= -1
    [[Qxx, Qxy, Qxz], [Qyx, Qyy, Qyz], [Qzx, Qzy, Qzz]] = mat
    K = np.asarray([
        [Qxx - Qyy - Qzz, Qyx + Qxy, Qzx + Qxz, Qzy - Qyz],
        [Qyx + Qxy, Qyy - Qxx - Qzz, Qzy + Qyz, Qxz - Qzx],
        [Qzx + Qxz, Qzy + Qyz, Qzz - Qxx - Qyy, Qyx - Qxy],
        [Qzy - Qyz, Qxz - Qzx, Qyx - Qxy, Qxx + Qyy + Qzz],
    ]) / 3
    val, vec = np.linalg.eig(K)
    vec = vec[:, val.argmax()]
    # vec = np.concatenate([vec[1:], vec[:1]])  # wiki to neuroglancer
    return vec


def quaternion_to_matrix(q: np.ndarray) -> np.ndarray:
    """Convert a quaternion to a rotation matrix."""
    # https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#From_a_quaternion_to_an_orthogonal_matrix
    q = np.asarray(q).flatten()
    if q.shape != (4,):
        raise ValueError("Expected a vector of length 4")
    q = np.concatenate([q[-1:], q[:-1]])  # neuroglancer to wiki
    a, b, c, d = q
    s = 2 / (a*a + b*b + c*c + d*d)
    bs = b * s
    cs = c * s
    ds = d * s
    ab = a * bs
    ac = a * cs
    ad = a * ds
    bb = b * bs
    bc = b * cs
    bd = b * ds
    cc = c * cs
    cd = c * ds
    dd = d * ds
    m = np.asarray([
        [1 - cc - dd, bc - ad, bd + ac],
        [bc + ad, 1 - bb - dd, cd - ab],
        [bd - ac, cd + ab, 1 - bb - cc],
    ])
    return m


def compose_quaternions(a: np.ndarray, b: np.ndarray, *others) -> np.ndarray:
    """Compose quaternions."""
    a = np.asarray(a).flatten()
    b = np.asarray(b).flatten()
    av, ar = a[:-1], a[-1:]
    bv, br = b[:-1], b[-1:]
    cr = ar * br - np.dot(av, bv)
    cv = ar * bv + br * av + np.cross(av, bv)
    c = np.concatenate([cv, cr])
    if others:
        c = compose_quaternions(c, *others)
    return c


def inverse_quaternions(a: np.ndarray) -> np.ndarray:
    """Inverse quaternion."""
    a = np.asarray(a).flatten()
    av, ar = a[:-1], a[-1:]
    nrm = ar*ar + (av*av).sum()
    ar = +ar / nrm
    av = -av / nrm
    a = np.concatenate([av, ar])
    return a


def ras2transform(
    ras2ras: np.ndarray,
    names: list[str] = ("x", "y", "z"),
) -> ng.CoordinateSpaceTransform:
    """Convert a RAS2RAS matrix (in mm space) to a neuroglancer transform."""
    dims = ng.CoordinateSpace(names=names, scales=[1]*3, units=["mm"]*3)
    return ng.CoordinateSpaceTransform(
        matrix=ras2ras[:3],
        input_dimensions=dims,
        output_dimensions=dims,
    )


def load_affine(
    fileobj: IO | PathLike | str,
    format: str | None = None,
    moving: PathLike | str | None = None,
    fixed: PathLike | str | None = None,
    names: list[str] = ("x", "y", "z"),
) -> ng.CoordinateSpaceTransform:
    """
    Load an affine transform from a file.

    Parameters
    ----------
    fileobj : file_like or str
        Input file. May start with a format protocol such as
        `"afni://path/to/file"`. Accepeted protocols are

        * `"afni://"`   Affine generated by AFNI
        * `"fsl://"`    Affine generated by FSL FLIRT
        * `"itk://"`    Affine in ITK format (e.g., generated by ANTs)
        * `"lta://"`    Affine in FreeSurfer LTA format (mri_robust_register)
    format : {"afni", "fsl", "itk", "lta"}, optional
        Alternative way to provide a format hint.

    Returns
    -------
    affine : (4, 4) np.ndarray
        A RAS2RAS affine transform.
    """
    opt = dict(moving=moving, fixed=fixed, names=names)

    fileobj = stringify_path(fileobj)

    if isinstance(fileobj, str):
        parsed = parse_protocols(fileobj)
        format = format or parsed.format
        fileobj = parsed.url

        if not format and fileobj.endswith('.lta'):
            format = 'lta'

        # If remote file, open with fsspec
        if parsed.stream != "file":
            with open(fileobj) as f:
                return load_affine(f, format=format, **opt)

    # If open file object, write to local file
    if hasattr(fileobj, 'read'):
        with TemporaryDirectory() as tmp:
            # NOTE: I use a TemporaryDirectory rather than a
            #       NamedTemporaryFile because I've encountered
            #       issues where the file was not yet written
            #       but was already being read by the transform reader.
            tmpname = Path(tmp) / "affine.lta"
            with tmpname.open("wb") as f:
                f.write(fileobj.read())
            out = load_affine(tmpname, format=format, **opt)
            return out

    def _read(klass: type) -> ng.CoordinateSpaceTransform:
        """Try to read with a nitransforms type."""
        ras2ras = klass.from_filename(fileobj).to_ras(moving, fixed)
        if isinstance(ras2ras, list):
            ras2ras = ras2ras[0]
        # NOTE:
        #   * nitransforms seems to think it should invert the matrix stored
        #     in the LTA, but I don't follow their motivation, so I uninvert
        #     it.
        ras2ras = np.linalg.inv(ras2ras)
        return _make_transform(ras2ras)

    def _make_transform(ras2ras: np.ndarray) -> ng.CoordinateSpaceTransform:
        dims = ng.CoordinateSpace(names=names, scales=[1]*3, units=["mm"]*3)
        return ng.CoordinateSpaceTransform(
            matrix=ras2ras[:3],
            input_dimensions=dims,
            output_dimensions=dims,
        )

    if format == "lta":
        # Try my reader first
        # (nitransforms does not like nitorch's cropped LTAs)
        try:
            out = LinearTransformArray(fileobj).matrix()[0]
            out = _make_transform(out)
            LOG.debug(f'Succesfully read format "{format}".')
            return out
        except Exception as e:
            LOG.info(f'Tried format "{format}" (our impl) with no success.')
            LOG.debug(e)
            raise e

    if format in AFFINE_FORMATMAP:
        try:
            out = _read(AFFINE_FORMATMAP[format])
            LOG.debug(f'Succesfully read format "{format}".')
            return out
        except Exception as e:
            LOG.warning(f'Tried format "{format}" with no success.')
            LOG.debug(e)

    hint_format = format
    for format, klass in AFFINE_FORMATMAP.items():
        if format == hint_format:
            continue
        try:
            out = _read(klass)
            LOG.info(f'Succesfully read format "{format}".')
            return out
        except Exception as e:
            LOG.info(f'Tried format "{format}" with no success.')
            LOG.debug(e)

    if hint_format != "lta":
        # Try my reader last
        # (nitransforms does not like nitorch's cropped LTAs)
        try:
            out = LinearTransformArray(fileobj).matrix()[0]
            LOG.debug(f'Succesfully read format "{format}".')
            return out
        except Exception as e:
            LOG.warning(f'Tried format "{format}" (our impl) with no success.')
            LOG.debug(e)

    LOG.error(f"Failed to load {fileobj}.")
    raise RuntimeError(f"Failed to load {fileobj}.")


def save_transform(
    trf: ng.CoordinateSpaceTransform | dict,
    out: str,
    format: str | None = None,
) -> None:
    """
    Save a neuroglancer transform in a neuroimaging format.

    Parameters
    ----------
    trf : ng.CoordinateSpaceTransform | dict
        Transformation
    out : str | PathLike
        Output path. Currently, only local paths are supported.
    format : str | None
        Output format (default: guessed from extension -- or LTA)
    """
    # ensure CoordinateSpaceTransform
    if isinstance(trf, np.ndarray):
        # assume millimetric RAS2RAS
        if trf.shape[0] == trf.shape[1]:
            trf = trf[:-1]
        names = ["x", "y", "z"]
        dims = ng.CoordinateSpace(names=names, scales=[1]*3, units=["mm"]*3)
        trf = ng.CoordinateSpaceTransform(
            matrix=trf,
            input_dimensions=dims,
            output_dimensions=dims,
        )

    trf = ng.CoordinateSpaceTransform(trf)
    trf = convert_transform(trf, [1, "mm"], [1, "mm"])
    idims = trf.input_dimensions
    odims = trf.output_dimensions or idims
    matrix = get_matrix(trf, square=True)

    iperm = [idims.names.index(x) for x in 'xyz']
    operm = [odims.names.index(x) for x in 'xyz']
    matrix = matrix[operm + [-1], :][:, iperm + [-1]]

    # NOTE:
    #   * nitransforms seems to think it should invert the matrix stored
    #     in the LTA, but I don't follow their motivation, so I uninvert
    #     it.
    matrix = np.linalg.inv(matrix)

    out = str(out)
    if format is None:
        for maybe_format in AFFINE_FORMATMAP:
            if out.endswith("." + maybe_format):
                format = maybe_format

    if format:
        try:
            kls = AFFINE_FORMATMAP[format]
            kls.from_ras(matrix).to_filename(out)
            LOG.info(f'Succesfully saved to format "{format}".')
            return
        except Exception as e:
            LOG.info(f'Tried format "{format}" with no success.')
            LOG.debug(e)

    hint_format = format
    for format, kls in AFFINE_FORMATMAP.items():
        if format == hint_format:
            continue
        try:
            kls = AFFINE_FORMATMAP[format]
            kls.from_ras(matrix).to_filename(out)
            LOG.info(f'Succesfully saved to format "{format}".')
            return out
        except Exception as e:
            LOG.info(f'Tried format "{format}" with no success.')
            LOG.debug(e)

    LOG.error(f"Failed to save {out}.")
    raise RuntimeError(f"Failed to save {out}.")


def to_square(affine: np.ndarray) -> np.ndarray:
    """Ensure an affine matrix is in square (homogeneous) form."""
    if affine.shape[0] == affine.shape[1]:
        return affine
    new_affine = np.eye(affine.shape[-1])
    new_affine[:-1, :] = affine
    return new_affine


def get_matrix(
    trf: ng.CoordinateSpaceTransform,
    square: bool = False
) -> np.ndarray:
    """Get (a copy of) the transfomation's matrix -- create one if needed."""
    rank = len(trf.input_dimensions.names)
    matrix = trf.matrix
    if matrix is None:
        matrix = np.eye(rank+1)[:-1]
    if square:
        matrix = to_square(np.copy(matrix))
    if not square and matrix.shape[0] == matrix.shape[1]:
        matrix = matrix[:-1]
    return matrix


def inverse(
    transform: ng.CoordinateSpaceTransform
) -> ng.CoordinateSpaceTransform:
    """Invert a transform."""
    # Invert normalized transform
    #
    #   In normalized units we have
    #
    #       [out] = linear @ [inp] + out_scale @ [offset]
    #
    #   yielding the inverse
    #
    #       [inp] = inv(linear) @ [out] - inv(linear) @ out_scale @ [offset]
    #
    #   Assuming that we reuse the same input/output dimensions, but swapped,
    #   we build the inverse transform by
    #
    #       linear = inv(linear)
    #       offset = - inv(inp_scale) @ inv(linear) @ out_scale @ [offset]
    #
    #   In our case, we've already ensured that inp_scale == out_scale == 1.

    matrix = get_matrix(transform, square=False)
    offset = matrix[:, -1:]
    linear = matrix[:, :-1]

    inp_scales = np.asarray([
        U.normalize_unit(x.scale, x.unit)
        for x in transform.input_dimensions
    ])[:, None]
    out_scales = np.asarray([
        U.normalize_unit(x.scale, x.unit)
        for x in transform.output_dimensions
    ])[:, None]

    inv_linear = np.linalg.inv(linear)
    inv_offset = -(inv_linear @ (offset * out_scales)) / inp_scales
    inv_matrix = np.concatenate([inv_linear, inv_offset], -1)

    inv_transform = ng.CoordinateSpaceTransform(
        input_dimensions=transform.output_dimensions,
        output_dimensions=transform.input_dimensions,
        matrix=inv_matrix
    )
    return inv_transform


def compose(
    *transforms,
    adapt: bool = False,
) -> ng.CoordinateSpaceTransform:
    """
    Compose two transforms.

    Parameters
    ----------
    *transforms : ng.CoordinateSpaceTransform
        Transforms to compose
    adapt : bool
        Try to adapt midspaces if they are different neurospaces.

    Returns
    -------
    composition : ng.CoordinateSpaceTransform
    """
    if len(transforms) == 0:
        return None
    if len(transforms) == 1:
        return transforms[0]
    left, right, *transforms = transforms
    if transforms:
        opt = dict(adapt=adapt)
        return compose(compose(left, right, **opt), *transforms, **opt)

    # ------------------------------------------------------------------
    # Adapt midspace
    #
    #   The left and right transforms may map from/to spaces that are
    #   compatible but different (for example, different neurospaces
    #   such as "ras" and "lpi"). We therefore convert the intermediate
    #   spaces (right output and left input) to the same neurospace
    #   (xyz == ras) first.
    if adapt:
        space_right = S.space_to_name(right.output_dimensions, compact=True)
        space_left = S.space_to_name(left.input_dimensions, compact=True)
        sorted_right = list(sorted(space_right))
        sorted_left = list(sorted(space_left))
        if sorted_right != ["x", "y", "z"][:len(sorted_right)]:
            right = compose(S.neurotransforms[("xyz", space_right)], right)
        if sorted_left != ["x", "y", "z"][:len(sorted_left)]:
            left = compose(left, S.neurotransforms[(space_left, "xyz")])

    # ------------------------------------------------------------------
    # Get parts
    li_dims = left.input_dimensions.to_json()
    lo_dims = left.output_dimensions.to_json()
    ri_dims = right.input_dimensions.to_json()
    ro_dims = right.output_dimensions.to_json()

    l_matrix = get_matrix(left, square=True)
    r_matrix = get_matrix(right, square=True)

    # ------------------------------------------------------------------
    # Filter left transform
    #
    #   Any input axis of the left transform that does not correspond
    #   to an output axis of the right transform must be removed.
    #
    #   This is because it is not possible to insert an dimension that
    #   did not exist in the input tensor in neuroglancer (AFAIK!).
    #
    #   However, the dimension that we drop should not be mixed in
    #   any way with the dimensions that are preserved. To check this we:
    #
    #       1. Find all output dimensions that depend (have nonzero weight)
    #          on one of the deleted dimensions.
    #       2. Check that each of these output dimensions _only_ depends
    #          (has non zero weight) on deleted dimensions.
    #
    #   In other words, if an output axis depends on both a deleted and
    #   a preserved axis, we have a problem.

    li_names = list(li_dims.keys())
    lo_names = list(lo_dims.keys())

    # List of input names to remove from left transform
    li_names_rm = [name for name in li_dims if name not in ro_dims]
    li_indcs_rm = [li_names.index(name) for name in li_names_rm]

    # Check mixing across removed and preserved axes
    lo_indcs_rm = set()
    for li_indx in li_indcs_rm:
        lo_indces = l_matrix[:-1, li_indx].nonzero()[0].tolist()
        for lo_indx in lo_indces:
            li_indces = l_matrix[lo_indx, :-1].nonzero()[0].tolist()
            if set(li_indces) - set(li_indcs_rm):
                raise RuntimeError(
                    "Transforms are not compatible. The left transform "
                    "mixes axes that are present in the right transform "
                    "with axes that are missing from the right transform."
                )
        lo_indcs_rm |= set(lo_indces)
    lo_names_rm = [lo_names[indx] for indx in lo_indcs_rm]

    # Delete rows and columns from matrix
    li_keep = [i for i in range(len(li_names)) if i not in li_indcs_rm]
    lo_keep = [i for i in range(len(lo_names)) if i not in lo_indcs_rm]
    l_matrix = l_matrix[li_keep + [-1], :][:, lo_keep + [-1]]

    # Filter dimensions
    li_dims = dict(filter(lambda x: x[0] not in li_names_rm, li_dims.items()))
    lo_dims = dict(filter(lambda x: x[0] not in lo_names_rm, lo_dims.items()))

    # ------------------------------------------------------------------
    # Populate left transform
    #
    #   We also need to insert into the left transform an identity
    #   sub-transform that corresponds to axes that are present in the
    #   output dimensions of the right axis but not in the input dimensions
    #   of the left transform.

    li_names = list(li_dims.keys())
    ro_names = list(ro_dims.keys())

    # Find axes that exist in the input space but not in the output space
    li_names_add = [name for name in ro_dims if name not in li_dims]

    # Check that these axis names do not already exist in the ouput
    # dimensions of the right matrix.
    if any(name in lo_dims for name in li_names_add):
        raise RuntimeError(
            "Transforms are not compatible. The left transform defines "
            "an output axis whose name conflicts with an output axis "
            "of the right transform."
        )

    # Add axes to the input/output dimensions of the right transform
    li_dims.update({name: ro_dims[name] for name in li_names_add})
    lo_dims.update({name: ro_dims[name] for name in li_names_add})

    # Insert identity matrix in the left matrix
    l_ndims = len(l_matrix) - 1
    r_ndims = len(r_matrix) - 1
    l_matrix0, l_matrix = l_matrix, np.eye(r_ndims+1)
    # -> copy compatible part
    l_matrix[:l_ndims, :l_ndims] = l_matrix0[:-1, :-1]
    l_matrix[:l_ndims, -1:] = l_matrix0[:-1, -1:]

    # Reorder right side of the matrix to conform to the right matrix
    li_names = list(li_dims.keys())
    l_matrix = l_matrix[:, [li_names.index(n) for n in ro_names] + [-1]]

    # ------------------------------------------------------------------
    # Compose matrices
    #
    #   A voxel to world matrix is constructed from a ng transform by:
    #
    #   [ x ]   [ 1/vx     ]   [ Axi Axj Axk ]   [ sx       ]   [ i ]   [ Tx ]
    #   [ y ] = [   1/vy   ] * [ Ayi Ayj Ayk ] * [    sy    ] * [ j ] + [ Ty ]
    #   [ z ]   [     1/vz ]   [ Azi Azj Azk ]   [       sz ]   [ k ]   [ Tz ]
    #
    #   where `v` is the view/output voxel size and `s` is the source/input
    #   voxel size. In this context, `[x, y, z]` is in "view" scaled unit and
    #   `[i, j, k]` is in "source" scaled unit.
    #
    #   If we're working in unscaled SI units, we instead have
    #
    #   [ x ]   [ Axi Axj Axk ]   [ i ]   [ vx       ]   [ Tx ]
    #   [ y ] = [ Ayi Ayj Ayk ] * [ j ] + [    vy    ] * [ Ty ]
    #   [ z ]   [ Azi Azj Azk ]   [ k ]   [       vz ]   [ Tz ]
    #
    #   where both `[x, y, z]` and `[i, j, k]` are in unscaled SI units
    #   (i.e., "m" or "s").
    #
    #   To compose two transforms
    #
    #   [mid] = rlinear @ [inp] + rout_scale @ [roffset]
    #   [out] = llinear @ [mid] + lout_scale @ [loffset]
    #
    #   where `out`, `inp` and `mid` are in unscaled SI units,
    #   one therefore does
    #
    #   [out] = llinear @ rlinear    @ [inp]
    #         + llinear @ rout_scale @ [roffset]
    #         +           lout_scale @ [loffset]
    #
    #   assuming the the input scale of the composed transform is `rinp_scale`
    #   and its output scale is `lout_scale`, the matrix is therefore
    #   constructed by
    #
    #   linear = llinear @ rlinear
    #   offset = loffset + inv(lout_scale) @ llinear @ rout_scale @ roffset

    l_linear = l_matrix[:-1, :-1]
    r_linear = r_matrix[:-1, :-1]
    l_offset = l_matrix[:-1, -1:]
    r_offset = r_matrix[:-1, -1:]

    ro_scles = np.asarray([
        U.normalize_unit(scale, unit)
        for (scale, unit) in ro_dims.values()
    ])[:, None]
    lo_scles = np.asarray([
        U.normalize_unit(scale, unit)
        for (scale, unit) in lo_dims.values()
    ])[:, None]

    matrix = np.eye(len(l_matrix))
    matrix[:-1, :-1] = l_linear @ r_linear
    matrix[:-1, -1:] = l_offset + (l_linear @ (ro_scles * r_offset)) / lo_scles

    return ng.CoordinateSpaceTransform(
        input_dimensions=ng.CoordinateSpace(ri_dims),
        output_dimensions=ng.CoordinateSpace(lo_dims),
        matrix=matrix[:-1],
    )


AxisConversionSpecification = (
    ng.DimensionScale |
    str | list[str] |                               # unit(s)
    tuple[float, str] | list[tuple[float, str]]     # scale(s) + unit(s)
)
SpaceConversionSpecification = (
    ng.CoordinateSpace |
    AxisConversionSpecification |
    dict[str, AxisConversionSpecification]
)


def convert_transform(
    transform: ng.CoordinateSpaceTransform,
    input: SpaceConversionSpecification | None = None,
    output: SpaceConversionSpecification | None = None,
) -> ng.CoordinateSpaceTransform:
    """
    Convert units of a transform.

    !!! example
        === "A"
            __Ensure that all axes have SI unit (without prefix)__
            ```python
            new_space = convert_transform(space, "base", "base")
            # or
            base = ["m", "s", "hz", "rad/s"]
            new_space = convert_space(space, base, base)
            # or
            new_space = normalize_transform(space)
            ```
        === "B"
            __Ensure that all axes have unit-scale SI unit (without prefix)__
            ```python
            base = [[1, "m"], [1, "s"], [1, "hz"], [1, "rad/s"]]
            new_space = convert_space(space, base, base)
            # or
            new_space = normalize_transform(space, unit_scale=True)
            ```
        === "C"
            __Ensure that input spatial axes have millimeter unit__
            ```python
            new_space = convert_space(space, "mm")
            # or
            new_space = convert_space(space, "mm", None)
            ```
        === "D"
            __Ensure that all spatial axes have millimeter unit__
            ```python
            new_space = convert_space(space, "mm", "mm")
            ```
        === "E"
            __Ensure that specific input axes have specific unit__
            ```python
            new_space = convert_space(space, {"x": "mm", "y": "mm", "t": "ms"})
            # or
            new_space = convert_space(space, {("x", "y"): "mm", "t": "ms"})
            # or
            new_space = convert_space(space, {("x", "y", "t"): ["mm", "ms"]})
            ```
        === "F"
            __Ensure that specific input axes have specific scales__
            ```python
            new_space = convert_space(space, {"x": [2, "mm"], "y": [2, "mm"], "t": [3, "ms"]})
            # or
            new_space = convert_space(space, {("x", "y"): [2, "mm"], "t": [3, "ms"]})
            # or
            new_space = convert_space(space, {("x", "y", "t"): [[2, "mm"], [3, "ms"]]})
            # or
            new_space = convert_space(space,
            ```

    !!! info "See also: `normalize_transform`."

    Parameters
    ----------
    transform : ng.CoordinateSpaceTransform
        Transform
    input : ng.CoordinateSpace | str | tuple[float, str] | list | dict | None
        Either a coordinate space to use in-place of the existing one,
        or a new (list of) unit(s), or a mapping from axis name(s) to
        new unit(s).
    output : ng.CoordinateSpace | str | tuple[float, str] | list | dict | None
        Either a coordinate space to use in-place of the existing one,
        or a new (list of) unit(s), or a mapping from axis name(s) to
        new unit(s).

    Returns
    -------
    transform : ng.CoordinateSpaceTransform
    """  # noqa: E501
    rank = len(transform.input_dimensions.names)
    old_input: ng.CoordinateSpace = transform.input_dimensions
    old_output: ng.CoordinateSpace = transform.output_dimensions

    def get_spec_unit(
        spec: SpaceConversionSpecification
    ) -> dict[str, float]:
        """Extract the unit part of the specification."""
        if isinstance(spec, (list, tuple)) \
                and not isinstance(spec, ng.DimensionScale):
            if isinstance(spec[0], Number):
                spec = ng.DimensionScale.from_json(list(spec))
            else:
                return [get_spec_unit(spec1) for spec1 in spec]
        if isinstance(spec, ng.DimensionScale):
            return spec.unit
        if isinstance(spec, str):
            return spec
        if isinstance(spec, dict):
            return {
                name1: get_spec_unit(spec1)
                for name1, spec1 in spec.items()
            }
        raise TypeError("Unknown specification type:", type(spec))

    def apply_spec_scale(
        space: ng.CoordinateSpace,
        spec: SpaceConversionSpecification,
        names: list[str] | None = None,
    ) -> ng.CoordinateSpace:
        """Apply the scaling part of the specification."""
        if names is None:
            names = space.names
        if isinstance(spec, (list, tuple)) \
                and not isinstance(spec, ng.DimensionScale):
            if isinstance(spec[0], Number):
                spec = ng.DimensionScale.from_json(list(spec))
            else:
                for spec1 in spec:
                    space = apply_spec_scale(space, spec1)
                return space
        if isinstance(spec, ng.DimensionScale):
            space = space.to_json()
            for name in names:
                if U.same_unit_kind(space[name][1], spec.unit):
                    space[name] = [spec.scale, spec.unit]
            space = ng.CoordinateSpace(space)
            return space
        if isinstance(spec, str):
            return space
        if isinstance(spec, dict):
            for name, spec1 in spec.items():
                space = apply_spec_scale(space, spec1, _ensure_list(name))
            return space
        raise TypeError("Unknown specification type:", type(spec))

    input = input or old_input
    if not isinstance(input, ng.CoordinateSpace):
        spec = input
        old_input = ng.CoordinateSpace(old_input.to_json())  # copy
        input = S.convert_space(old_input, get_spec_unit(spec))
        input = apply_spec_scale(input, spec)

    output = output or old_output
    if not isinstance(output, ng.CoordinateSpace):
        spec = output
        old_output = ng.CoordinateSpace(old_output.to_json())  # copy
        output = S.convert_space(old_output, get_spec_unit(spec))
        output = apply_spec_scale(output, spec)

    matrix = get_matrix(transform, square=True)

    for i in range(rank):
        # input (scale columns of the linear component)
        old_scale = old_input.scales[i]
        new_scale = input.scales[i]
        old_unit = old_input.units[i]
        new_unit = input.units[i]
        old_scale = U.convert_unit(old_scale, old_unit, new_unit)
        matrix[:-1, i] *= (old_scale / new_scale)
        # output (scale the translation)
        old_scale = old_output.scales[i]
        old_unit = old_output.units[i]
        new_scale = output.scales[i]
        new_unit = output.units[i]
        old_scale = U.convert_unit(old_scale, old_unit, new_unit)
        matrix[i, -1] *= (old_scale / new_scale)

    return ng.CoordinateSpaceTransform(
        input_dimensions=input,
        output_dimensions=output,
        matrix=matrix[:-1, :],
    )


def normalize_transform(
    transform: ng.CoordinateSpaceTransform,
    unit_scale: bool = False,
) -> ng.CoordinateSpaceTransform:
    """Ensure that input and output spaces have SI units.

    !!! info "See also: `convert_transform`, `ensure_same_scale`."

    Parameters
    ----------
    transform : ng.CoordinateSpaceTransform
        Transformation
    unit_scale : bool
        If `True`, make input and output space to have unit scale.
        Otherwise, simply convert them to SI unit.

    Returns
    -------
    transform : ng.CoordinateSpaceTransform
        Converted transformation
    """
    spec = ["m", "s", "hz", "rad/s"]
    if unit_scale:
        spec = [(1, unit) for unit in spec]
    return convert_transform(transform, spec, spec)


def ensure_same_scale(
    transform: ng.CoordinateSpaceTransform
) -> ng.CoordinateSpaceTransform:
    """
    Ensure that all axes and both input and output spaces use the same
    unit and the same scaling.

    This function is quite related to `normalize_transform(*, unit_scale=True)`
    except that the units closest to input units are used, instead of
    prefixless SI units.

    !!! info "See also: `normalize_transform`."
    """
    # Find all scales and prefixes in the input/output spaces
    #   The `unitmap` dictionary will map unit kinds present in the
    #   transform's dimensions to a list of possible exponents.
    unitmap = {}
    for dimensions in (transform.outputDimensions, transform.inputDimensions):
        dimensions: ng.CoordinateSpace
        for unit, scale in zip(dimensions.units, dimensions.scales):
            prefix, kind = U.split_unit(unit)
            exponent = U.SI_PREFIX_EXPONENT[prefix]
            exponent += int(round(math.log10(scale)))
            unitmap.setdefault(kind, [])
            unitmap[kind].append(exponent)

    # Choose the unit and scale that are closest/most common
    #   The `scalemap` dictionary will map unit kinds present in the
    #   transform's dimensions to a list of (scale, unit) to convert to.
    scalemap = {}
    for kind, exponents in dict(unitmap).items():
        # find most common exponent
        exponent, count = None, 0
        for maybe_exponent in set(exponents):
            if exponents.count(maybe_exponent) > count:
                exponent = maybe_exponent
        # find SI prefix closest to exponent and distance to exponent
        prefix, dist = None, float("inf")
        for maybe_prefix, maybe_exponent in U.SI_PREFIX_EXPONENT.items():
            if abs(exponent - maybe_exponent) < abs(dist):
                dist = exponent - maybe_exponent
                prefix = maybe_prefix
        # map unit kinf
        scalemap[kind] = [10 ** dist, prefix + kind]

    # convert the transform
    for kind, scale in scalemap.items():
        spec = ng.DimensionScale(*scale)
        transform = convert_transform(transform, spec, spec)
    return transform


def subtransform(
    transform: ng.CoordinateSpaceTransform,
    unit: str | list[str] | None = None,
    *,
    names: str | list[str] | None = None,
    input_names: str | list[str] | None = None,
    output_names: str | list[str] | None = None,
) -> ng.CoordinateSpaceTransform:
    """
    Generate a subtransform by keeping only axes of a certain unit kind.

    Parameters
    ----------
    transform : ng.CoordinateSpaceTransform
        Transformation
    unit : str | list[str]
        Unit kind to keep.

    Other Parameters
    ----------------
    names : str | list[str], optional
        Axis names to keep.
    input_names, str | list[str], default=`names`
        Names of input axes to keep.
    output_names, str | list[str], default=`names`
        Names of output axes to keep.

    Returns
    -------
    transform : ng.CoordinateSpaceTransform
        Converted transformation

    """
    kind = None
    if unit is not None:
        unit = _ensure_list(unit)
        kind = [U.split_unit(u)[1] for u in unit]
    input_names = _ensure_list(input_names or names)
    output_names = _ensure_list(output_names or names)

    idim: ng.CoordinateSpace = transform.input_dimensions
    odim: ng.CoordinateSpace = transform.output_dimensions
    matrix = transform.matrix
    if matrix is None:
        matrix = np.eye(len(idim.names)+1)[:-1]
    # filter input axes
    ikeep = []
    for i, (name, unit) in enumerate(zip(idim.names, idim.units)):
        unit = U.split_unit(unit)[1]
        if kind and unit not in kind:
            continue
        if input_names and name not in input_names:
            continue
        ikeep.append(i)
    idim = ng.CoordinateSpace({
        idim.names[i]: [idim.scales[i], idim.units[i]]
        for i in ikeep
    })
    matrix = matrix[:, ikeep + [-1]]
    # filter output axes
    okeep = []
    for i, (name, unit) in enumerate(zip(odim.names, odim.units)):
        unit = U.split_unit(unit)[1]
        if kind and unit not in kind:
            continue
        if output_names and name not in output_names:
            continue
        okeep.append(i)
    odim = ng.CoordinateSpace({
        odim.names[i]: [odim.scales[i], odim.units[i]]
        for i in okeep
    })
    matrix = matrix[okeep, :]

    return ng.CoordinateSpaceTransform(
        matrix=matrix,
        input_dimensions=idim,
        output_dimensions=odim,
    )


def _ensure_list(x: object, n: int | None = None, **kwargs) -> list:
    if isinstance(x, np.ndarray):
        x = x.tolist()
    if not isinstance(x, (list, tuple)):
        x = [x]
    x = list(x)
    if n is not None:
        if len(x) < n:
            default = kwargs.get("default", x[-1])
            x += [default] * (n - len(x))
        elif len(x) > n:
            x = x[:n]
    return x


def _ensure_tuple(x: object, n: int | None = None, **kwargs) -> tuple:
    return tuple(_ensure_list(x, n, **kwargs))



# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Data structures for the X5 transform format.

Implements what's drafted in the `BIDS X5 specification draft
<https://docs.google.com/document/d/1yk5O0QTAOXLdP9iSG3W8ta7IcMFypu2106c3Pnjfi-4/edit>`__.

"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, List

import json
import h5py

import numpy as np


@dataclass(eq=True)
class X5Domain:
    """Domain information of a transform representing reference/moving spaces."""

    grid: bool
    """Whether sampling locations in the manifold are located regularly."""
    size: Sequence[int]
    """The number of sampling locations per dimension (or total if not a grid)."""
    mapping: Optional[np.ndarray]
    """A mapping to go from samples (pixel/voxel coordinates, indices) to space coordinates."""
    coordinates: Optional[str] = None
    """Indexing type of the Mapping field (for example, "cartesian", "spherical" or "index")."""


@dataclass
class X5Transform:
    """Represent one transform entry of an X5 file."""

    type: str
    """A REQUIRED unicode string with possible values: "linear", "nonlinear", "composite"."""
    transform: np.ndarray
    """A REQUIRED array of parameters (e.g., affine matrix, or dense displacements field)."""
    subtype: Optional[str] = None
    """An OPTIONAL extension of type to drive the interpretation of AdditionalParameters."""
    representation: Optional[str] = None
    """
    A string specifiying the transform representation or model, REQUIRED only for nonlinear Type.
    """
    metadata: Optional[Dict[str, Any]] = None
    """An OPTIONAL string (JSON) to embed metadata."""
    dimension_kinds: Optional[Sequence[str]] = None
    """Identifies what "kind" of information is represented by the samples along each axis."""
    domain: Optional[X5Domain] = None
    """
    A dataset specifying the reference manifold for the transform (either
    a regularly gridded 3D space or a surface/sphere).
    REQUIRED for nonlinear Type, RECOMMENDED for linear Type.
    """
    inverse: Optional[np.ndarray] = None
    """Placeholder to pre-calculated inverses."""
    jacobian: Optional[np.ndarray] = None
    """
    A RECOMMENDED data array to keep cached the determinant of Jacobian of the transform
    in case tools have calculated it.
    For parametric models it is generally possible to obtain it analytically, so this dataset
    could not be as useful in that case.
    """
    additional_parameters: Optional[np.ndarray] = None
    """
    An OPTIONAL field to store additional parameters, depending on the SubType of the
    transform.
    """
    array_length: int = 1
    """Undocumented field in the draft to enable a single transform group for 4D transforms."""


def to_filename(fname: str | Path, x5_list: List[X5Transform]):
    """
    Write a list of :class:`X5Transform` objects to an X5 HDF5 file.

    Parameters
    ----------
    fname : :obj:`os.PathLike`
        The file name (preferably with the ".x5" extension) in which transforms will be stored.
    x5_list : :obj:`list`
        The list of transforms to be stored in the output dataset.

    Returns
    -------
    fname : :obj:`os.PathLike`
        File containing the transform(s).

    """
    with h5py.File(str(fname), "w") as out_file:
        out_file.attrs["Format"] = "X5"
        out_file.attrs["Version"] = np.uint16(1)
        tg = out_file.create_group("TransformGroup")
        for i, node in enumerate(x5_list):
            g = tg.create_group(str(i))
            _write_x5_group(g, node)
    return fname


def from_filename(fname: str | Path) -> List[X5Transform]:
    """Read a list of :class:`X5Transform` objects from an X5 HDF5 file."""
    try:
        with h5py.File(str(fname), "r") as in_file:
            if in_file.attrs.get("Format") != "X5":
                raise TypeError("Input file is not in X5 format")

            tg = in_file["TransformGroup"]
            return [
                _read_x5_group(node)
                for _, node in sorted(tg.items(), key=lambda kv: int(kv[0]))
            ]
    except OSError as err:
        if "file signature not found" in err.args[0]:
            raise TypeError("Input file is not HDF5.")

        raise  # pragma: no cover


def _read_x5_group(node) -> X5Transform:
    x5 = X5Transform(
        type=node.attrs["Type"],
        transform=np.asarray(node["Transform"]),
        subtype=node.attrs.get("SubType"),
        representation=node.attrs.get("Representation"),
        metadata=json.loads(node.attrs["Metadata"])
        if "Metadata" in node.attrs
        else None,
        dimension_kinds=[
            k.decode() if isinstance(k, bytes) else k
            for k in node["DimensionKinds"][()]
        ],
        domain=None,
        inverse=np.asarray(node["Inverse"]) if "Inverse" in node else None,
        jacobian=np.asarray(node["Jacobian"]) if "Jacobian" in node else None,
        array_length=int(node.attrs.get("ArrayLength", 1)),
        additional_parameters=np.asarray(node["AdditionalParameters"])
        if "AdditionalParameters" in node
        else None,
    )

    if "Domain" in node:
        dgrp = node["Domain"]
        x5.domain = X5Domain(
            grid=bool(int(np.asarray(dgrp["Grid"]))),
            size=tuple(np.asarray(dgrp["Size"])),
            mapping=np.asarray(dgrp["Mapping"]),
            coordinates=dgrp.attrs.get("Coordinates"),
        )

    return x5


def _write_x5_group(g, node: X5Transform):
    """Write one :class:`X5Transform` element into an opened HDF5 group."""
    g.attrs["Type"] = node.type
    g.attrs["ArrayLength"] = node.array_length
    if node.subtype is not None:
        g.attrs["SubType"] = node.subtype
    if node.representation is not None:
        g.attrs["Representation"] = node.representation
    if node.metadata is not None:
        g.attrs["Metadata"] = json.dumps(node.metadata)
    g.create_dataset("Transform", data=node.transform)
    g.create_dataset("DimensionKinds", data=np.asarray(node.dimension_kinds, dtype="S"))
    if node.domain is not None:
        dgrp = g.create_group("Domain")
        dgrp.create_dataset("Grid", data=np.uint8(1 if node.domain.grid else 0))
        dgrp.create_dataset("Size", data=np.asarray(node.domain.size))
        dgrp.create_dataset("Mapping", data=node.domain.mapping)
        if node.domain.coordinates is not None:
            dgrp.attrs["Coordinates"] = node.domain.coordinates
    if node.inverse is not None:
        g.create_dataset("Inverse", data=node.inverse)
    if node.jacobian is not None:
        g.create_dataset("Jacobian", data=node.jacobian)
    if node.additional_parameters is not None:
        g.create_dataset("AdditionalParameters", data=node.additional_parameters)



"""Implementation of the matrix exponential and its derivative.

This implementation is based on John Ashburner's in SPM, which relies
on a Taylor approximation for both the exponential and its derivative.
Faster implementations that rely on scaling and squaring or Pade
approximations (as in scipy) could be used instead. This may be the
object of future work.

"""
# externals
import numpy as np


def expm(X, basis=None, grad_X=False, grad_basis=False, hess_X=False,
         max_order=10000, tol=1e-32):
    """Matrix exponential (and its derivatives).

    Parameters
    ----------
    X : {(..., F), (..., D, D)} array
        If `basis` is None: log-matrix.
        Else:               parameters of the log-matrix in the basis set.
    basis : (..., F, D, D) array, default=None
        Basis set. If None, basis of all DxD matrices and F = D**2.
    grad_X : bool, default=False
        Compute derivatives with respect to `X`.
    grad_basis : bool, default=False
        Compute derivatives with respect to `basis`.
    max_order : int, default=10000
        Order of the Taylor expansion
    tol : float, default=1e-32
        Tolerance for early stopping
        The criterion is based on the Frobenius norm of the last term of
        the Taylor series.

    Returns
    -------
    eX : (..., D, D) tensor
        Matrix exponential
    dX : (..., F, D, D) tensor, if `grad_X is True`
        Derivative of the matrix exponential with respect to the
        parameters in the basis set
    dB : (..., F, D, D, D, D) tensor, if `grad_basis is True`
        Derivative of the matrix exponential with respect to the basis.
    hX : (..., F, F, D, D) tensor, if `hess_X is True`
        Second derivative of the matrix exponential with respect to the
        parameters in the basis set

    """

    # Authors
    # -------
    # .. John Ashburner <j.ashburner@ucl.ac.uk> : original Matlab code
    # .. Mikael Brudfors <brudfors@gmail.com> : Python port
    # .. Yael Balbastre <yael.balbastre@gmail.com> : Python port
    #
    # License
    # -------
    # The original Matlab code is (C) 2012-2019 WCHN / John Ashburner
    # and was distributed as part of [SPM](https://www.fil.ion.ucl.ac.uk/spm)
    # under the GNU General Public Licence (version >= 2).

    def smart_incr(a, b):
        """Use inplace or outplace increment based on `a` and `b`s shapes."""
        if a.shape == b.shape:
            a += b
        else:
            a = a + b
        return a

    X = np.asanyarray(X)

    if basis is not None:
        # X contains parameters in the Lie algebra -> reconstruct the matrix
        # X.shape = [.., F], basis.shape = [..., F, D, D]
        basis = np.asanyarray(basis, dtype=X.dtype)
        param = X
        X = np.sum(basis * X[..., None, None], axis=-3, keepdims=True)
        dim = basis.shape[-1]
    else:
        # X contains matrices in log-space -> build basis set
        # X.shape = [..., D, D]
        dim = X.shape[-1]
        param = X.reshape(X.shape[:-2] + (-1,))
        basis = np.eye(dim * dim, dtype=X.dtype)
        basis = basis.reshape([dim**2, dim, dim])
        X = X[..., None, :, :]

    XB_batch_shape = X.shape[:-3]
    nb_feat = param.shape[-1]

    if grad_basis:
        # Build a basis for the basis
        basis_basis = np.eye(dim * dim, dtype=X.dtype)
        basis_basis = basis_basis.reshape([1, dim, dim, dim, dim])
        basis_basis = basis_basis * param[..., None, None, None, None]
        basis_basis = basis_basis.reshape(XB_batch_shape + (-1, dim, dim))

    # At this point:
    #   X.shape           = [*XB_batch_shape, 1, D, D]
    #   basis.shape       = [*B_batch_shape,  F, D, D]
    #   param.shape       = [*X_batch_shape,  F]
    #   basis_basis.shape = [*XB_batch_shape, F*D*D, D, D]

    # Aliases
    I = np.eye(dim, dtype=X.dtype)
    E = I + X                            # expm(X)
    En = np.copy(X)                      # n-th Taylor coefficient of expm
    if grad_X or hess_X:
        dE = np.copy(basis)              # dexpm(X)/dx
        dEn = np.copy(basis)             # n-th Taylor coefficient of dexpm/dx
    if grad_basis:
        dB = np.copy(basis_basis)        # dexpm(X)/dB
        dBn = np.copy(basis_basis)       # n-th Taylor coefficient of dexpm/dB
    if hess_X:
        hE = np.zeros_like(E, shape=[*XB_batch_shape, nb_feat, nb_feat, dim, dim])
        hEn = np.copy(hE)

    for n_order in range(2, max_order+1):
        # Compute coefficients at order `n_order`, and accumulate
        if hess_X:
            dEB = np.matmul(dEn[..., None, :, :, :], basis[..., None, :, :])
            hEn = np.matmul(hEn, X[..., None, :, :]) + dEB + dEB.swapaxes(-3, -4)
            hEn /= n_order
            hE = smart_incr(hE, hEn)
            del dEB
        if grad_X:
            dEn = np.matmul(dEn, X) + np.matmul(En, basis)
            dEn /= n_order
            dE = smart_incr(dE, dEn)
        if grad_basis:
            dBn = np.matmul(dBn, X) + np.matmul(En, basis_basis)
            dBn /= n_order
            dB = smart_incr(dB, dBn)
        En = np.matmul(En, X)
        En /= n_order
        E = smart_incr(E, En)
        # Compute sum-of-squares
        sos = np.sum(En * En)
        if sos <= En.size * tol:
            break

    E = E[..., 0, :, :]
    if grad_basis:
        dB = dB.reshape(XB_batch_shape + (nb_feat, dim, dim, dim, dim))

    out = [E]
    if grad_X:
        out.append(dE)
    if grad_basis:
        out.append(dB)
    if hess_X:
        out.append(hE)
    return out[0] if len(out) == 1 else tuple(out)



# @license
# Copyright 2020 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for controlling a browser that runs Neuroglancer."""

import contextlib
import re
import sys
import threading
import time
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from selenium.webdriver.common.bidi.script import ConsoleLogEntry
else:
    ConsoleLogEntry = None

LogListener = Callable[[ConsoleLogEntry], None]


class WebdriverBase:
    def __init__(
        self,
        headless=True,
        browser="chrome",
        browser_binary_path: str | None = None,
        window_size=(1920, 1080),
        debug=False,
        docker=False,
        print_logs=True,
        extra_command_line_args: Sequence[str] | None = None,
    ):
        self.headless = headless
        self.browser = browser
        self.window_size = window_size
        self.headless = headless
        self.docker = docker
        self.extra_command_line_args = (
            list(extra_command_line_args) if extra_command_line_args else []
        )
        self.debug = debug
        self.browser_binary_path = browser_binary_path

        self._closed = False
        self._init_driver()

        if print_logs:
            self.add_log_listener(
                lambda log: print(f"console.{log.level}: {log.text}", file=sys.stderr)
            )

    def _init_chrome(self):
        import selenium.webdriver

        chrome_options = selenium.webdriver.ChromeOptions()
        chrome_options.enable_bidi = True
        if self.headless:
            chrome_options.add_argument("--headless=new")
        chrome_options.add_experimental_option("excludeSwitches", ["enable-automation"])
        if self.browser_binary_path:
            chrome_options.binary_location = self.browser_binary_path
        if self.docker:
            # https://www.intricatecloud.io/2019/05/running-webdriverio-tests-using-headless-chrome-inside-a-container/
            chrome_options.add_argument("--no-sandbox")
            chrome_options.add_argument("--disable-gpu")
            chrome_options.add_argument("--disable-setuid-sandbox")
            chrome_options.add_argument("--disable-dev-shm-usage")
        chrome_options.add_argument(
            "--window_size=%dx%d" % (self.window_size[0], self.window_size[1])
        )
        for arg in self.extra_command_line_args:
            chrome_options.add_argument(arg)
        self.driver = selenium.webdriver.Chrome(options=chrome_options)

    def _init_firefox(self):
        import selenium.webdriver

        options = selenium.webdriver.FirefoxOptions()
        if self.headless:
            options.add_argument("--headless")
        options.arguments.extend(self.extra_command_line_args)
        if self.browser_binary_path:
            options.binary_location = self.browser_binary_path
        options.enable_bidi = True
        self.driver = selenium.webdriver.Firefox(
            options=options,
        )

    def _init_driver(self):
        if self.browser == "chrome":
            self._init_chrome()
        elif self.browser == "firefox":
            self._init_firefox()
        else:
            raise ValueError(
                f'unsupported browser: {self.browser}, must be "chrome" or "firefox"'
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.driver.quit()
        self._closed = True

    def add_log_listener(self, listener: LogListener) -> Callable[[], None]:
        console_id = self.driver.script.add_console_message_handler(listener)
        error_id = self.driver.script.add_javascript_error_handler(listener)

        def unregister():
            self.driver.script.remove_console_message_handler(console_id)
            self.driver.script.remove_javascript_error_handler(error_id)

        return unregister

    @contextlib.contextmanager
    def log_listener(self, listener: LogListener):
        unregister = self.add_log_listener(listener)
        try:
            yield
        finally:
            unregister()

    @contextlib.contextmanager
    def wait_for_log_message(self, pattern: str, timeout: float | None = None):
        event = threading.Event()

        def handle_message(msg):
            if event.is_set():
                return
            if re.fullmatch(pattern, msg.text):
                event.set()

        with self.log_listener(handle_message):
            yield
            if not event.wait(timeout):
                raise TimeoutError

    def reload_browser(self):
        """Reloads the browser (useful if it crashes/becomes unresponsive)."""
        try:
            self.driver.quit()
        except Exception:
            pass
        self._init_driver()

    @property
    def root_element(self):
        return self.driver.find_element("xpath", "//body")

    def action_chain(self):
        import selenium.webdriver

        return selenium.webdriver.common.action_chains.ActionChains(self.driver)


class Webdriver(WebdriverBase):
    def __init__(self, viewer=None, **kwargs):
        if viewer is None:
            from .viewer import Viewer

            viewer = Viewer()
        self.viewer = viewer
        super().__init__(**kwargs)

    def _init_driver(self):
        super()._init_driver()
        self.driver.get(self.viewer.get_viewer_url())

    def sync(self):
        """Wait until client is ready."""
        while True:
            new_state = self.viewer.screenshot().viewer_state
            # Ensure self.viewer.state has also been updated to the new state.
            # The state sent in the screenshot reply can be newer.
            if new_state == self.viewer.state:
                return new_state
            time.sleep(0.1)



"""Read/write FSL's transforms."""
import os
import warnings
import numpy as np
from numpy.linalg import inv
from pathlib import Path
from nibabel.affines import voxel_sizes

from .base import (
    BaseLinearTransformList,
    LinearParameters,
    DisplacementsField,
    TransformIOError,
    TransformFileError,
    _ensure_image,
)


class FSLLinearTransform(LinearParameters):
    """A string-based structure for FSL linear transforms."""

    def __str__(self):
        """Generate a string representation."""
        lines = [
            " ".join("%.08f" % col for col in row)
            for row in self.structarr["parameters"]
        ]
        return "\n".join(lines + [""])

    def to_string(self):
        """Convert to a string directly writeable to file."""
        return self.__str__()

    @classmethod
    def from_ras(cls, ras, moving=None, reference=None):
        """Create an FSL affine from a nitransform's RAS+ matrix."""
        if moving is None:
            warnings.warn(
                "[Converting FSL to RAS] moving not provided, using reference as moving"
            )
            moving = reference

        if reference is None:
            raise TransformIOError("Cannot build FSL linear transform without a reference")

        reference = _ensure_image(reference)
        moving = _ensure_image(moving)

        # Adjust for reference image offset and orientation
        refswp, refspc = _fsl_aff_adapt(reference)
        pre = reference.affine @ inv(refswp @ refspc)

        # Adjust for moving image offset and orientation
        movswp, movspc = _fsl_aff_adapt(moving)
        post = movswp @ movspc @ inv(moving.affine)

        # Compose FSL transform
        mat = inv(np.swapaxes(post @ ras @ pre, 0, 1))

        tf = cls()
        tf.structarr["parameters"] = mat.T
        return tf

    @classmethod
    def from_string(cls, string):
        """Read the struct from string."""
        tf = cls()
        sa = tf.structarr
        lines = [line.strip() for line in string.splitlines() if line.strip()]
        if not lines or len(lines) < 4:
            raise TransformFileError

        sa["parameters"] = np.genfromtxt(
            ["\n".join(lines)], dtype=cls.dtype["parameters"]
        )
        return tf

    def to_ras(self, moving=None, reference=None):
        """Return a nitransforms internal RAS+ matrix."""
        if reference is None:
            raise TransformIOError("Cannot build FSL linear transform without a reference")

        if moving is None:
            warnings.warn(
                "Converting FSL to RAS: moving image not provided, using reference."
            )
            moving = reference

        reference = _ensure_image(reference)
        moving = _ensure_image(moving)

        refswp, refspc = _fsl_aff_adapt(reference)

        pre = refswp @ refspc @ inv(reference.affine)
        # Adjust for moving image offset and orientation
        movswp, movspc = _fsl_aff_adapt(moving)
        post = moving.affine @ inv(movswp @ movspc)
        mat = self.structarr["parameters"].T
        return post @ np.swapaxes(inv(mat), 0, 1) @ pre


class FSLLinearTransformArray(BaseLinearTransformList):
    """A string-based structure for series of FSL linear transforms."""

    _inner_type = FSLLinearTransform

    def to_filename(self, filename):
        """Store this transform to a file with the appropriate format."""
        output_dir = Path(filename).parent
        output_dir.mkdir(exist_ok=True, parents=True)
        for i, xfm in enumerate(self.xforms):
            (output_dir / f"{filename}.{i:03d}").write_text(str(xfm))

    def to_ras(self, moving=None, reference=None):
        """Return a nitransforms' internal RAS matrix."""
        return np.stack(
            [xfm.to_ras(moving=moving, reference=reference) for xfm in self.xforms]
        )

    def to_string(self):
        """Convert to a string directly writeable to file."""
        return "\n\n".join([xfm.to_string() for xfm in self.xforms])

    @classmethod
    def from_fileobj(cls, fileobj, check=True):
        """Read the struct from a file object."""
        return cls.from_string(fileobj.read())

    @classmethod
    def from_ras(cls, ras, moving=None, reference=None):
        """Create an ITK affine from a nitransform's RAS+ matrix."""
        _self = cls()
        _self.xforms = [
            cls._inner_type.from_ras(ras[i, ...], moving=moving, reference=reference)
            for i in range(ras.shape[0])
        ]
        return _self

    @classmethod
    def from_string(cls, string):
        """Read the struct from string."""
        _self = cls()
        _self.xforms = [cls._inner_type.from_string(string)]
        return _self

    @classmethod
    def from_filename(cls, filename):
        """
        Read the struct from a file given its path.

        If the file does not exist, then indexed names
        with the zero-padded suffix ``.NNN`` are
        attempted, following FSL's MCFLIRT conventions.

        """
        if os.path.exists(str(filename)):
            return super().from_filename(filename)

        _xforms = []
        index = 0
        while os.path.exists("%s.%03d" % (filename, index)):
            with open("%s.%03d" % (filename, index)) as f:
                string = f.read()
            _xforms.append(cls._inner_type.from_string(string))
            index += 1

        if index == 0:
            raise FileNotFoundError(str(filename))
        _self = cls()
        _self.xforms = _xforms
        return _self


class FSLDisplacementsField(DisplacementsField):
    """A data structure representing displacements fields."""

    @classmethod
    def from_image(cls, imgobj):
        """Import a displacements field from a NIfTI file."""
        hdr = imgobj.header.copy()
        shape = hdr.get_data_shape()

        if len(shape) != 4 or not shape[-1] in (2, 3):
            raise TransformFileError(
                'Displacements field "%s" does not come from FSL.' %
                imgobj.file_map['image'].filename)

        field = np.squeeze(np.asanyarray(imgobj.dataobj))
        field[..., 0] *= -1.0

        return imgobj.__class__(field, imgobj.affine, hdr)

    @classmethod
    def to_image(cls, imgobj):
        """Export a displacements field from a nibabel object."""

        hdr = imgobj.header.copy()

        warp_data = imgobj.get_fdata()
        warp_data[..., 0] *= -1

        return imgobj.__class__(warp_data, imgobj.affine, hdr)


def _fsl_aff_adapt(space):
    """
    Adapt FSL affines.

    Calculates a matrix to convert from the original RAS image
    coordinates to FSL's internal coordinate system of transforms
    """
    aff = space.affine
    zooms = list(voxel_sizes(aff)) + [1]
    swp = np.eye(4)
    if np.linalg.det(aff) > 0:
        swp[0, 0] = -1.0
        swp[0, 3] = (space.shape[0] - 1) * zooms[0]
    return swp, np.diag(zooms)



# @license
# Copyright 2023 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import tornado.httpserver
import tornado.netutil
import tornado.web

import neuroglancer.random_token
import neuroglancer.server

from . import async_util


class CorsStaticFileHandler(tornado.web.StaticFileHandler):
    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Headers", "x-requested-with, range")
        self.set_header("Access-Control-Expose-Headers", "content-range")
        self.set_header("Access-Control-Allow-Methods", "POST, GET, OPTIONS")

    def options(self, *args):
        self.set_status(204)
        self.finish()


class StaticFileServer(async_util.BackgroundTornadoServer):
    def __init__(
        self, static_dir: str, bind_address: str = "localhost", daemon=False
    ) -> None:
        super().__init__(daemon=daemon)
        self.bind_address = bind_address
        self.static_dir = static_dir

    def __enter__(self):
        return self.url

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()

    def _attempt_to_start_server(self):
        token = neuroglancer.random_token.make_random_token()
        handlers = [
            (rf"/{token}/(.*)", CorsStaticFileHandler, {"path": self.static_dir}),
        ]
        settings = {}
        self.app = tornado.web.Application(handlers, settings=settings)
        self.http_server = tornado.httpserver.HTTPServer(self.app)
        sockets = tornado.netutil.bind_sockets(port=0, address=self.bind_address)
        self.http_server.add_sockets(sockets)
        actual_port = sockets[0].getsockname()[1]
        server_url = neuroglancer.server._get_server_url(self.bind_address, actual_port)
        self.url = f"{server_url}/{token}"



# @license
# Copyright 2025 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reads annotations in the Precomputed annotation format."""

import collections
import json
import struct
import time
import typing

import numpy as np
import tensorstore as ts

from . import coordinate_space, viewer_state, write_annotations

K = typing.TypeVar("K")
V = typing.TypeVar("V")


def _get_uint64_key_encoder(metadata):
    if metadata is not None and "sharding" in metadata:
        return lambda n: struct.pack(">Q", n)
    return lambda n: str(n).encode("utf-8")


def _zorder_compressed(coords, grid_shape):
    zindex = 0
    output_bit = 0
    for bit in range(32):
        for coord, size in zip(coords, grid_shape):
            coord = int(coord)
            if (size - 1) >> bit:
                zindex |= ((coord >> bit) & 1) << output_bit
                output_bit += 1
    return zindex


def _get_spatial_key_encoder(metadata):
    if "sharding" in metadata:
        grid_shape = metadata["grid_shape"]
        return lambda coords: struct.pack(">Q", _zorder_compressed(coords, grid_shape))
    return lambda coords: "_".join(str(coord) for coord in coords).encode("utf-8")


class AnnotationMap(typing.Generic[K, V]):
    """Map interface for accessing an annotation index."""

    metadata: dict[str, typing.Any]
    """Raw JSON metadata associated with this annotation index."""

    def __init__(
        self,
        kvstore: ts.KvStore | None,
        metadata,
        key_encoder: typing.Callable[[K], bytes],
        value_decoder: typing.Callable[[K, bytes], V],
        staleness_bound: float | None = None,
    ):
        self._kvstore = kvstore
        self.metadata = metadata
        self._key_encoder = key_encoder
        self._value_decoder = value_decoder
        self.staleness_bound = staleness_bound

    def __bool__(self):
        return self._kvstore is not None

    def __getitem__(self, key: K) -> V:
        """Reads a given key."""
        if self._kvstore is None:
            raise KeyError("required index kind not available")
        read_result = self._kvstore.read(
            self._key_encoder(key), staleness_bound=self.staleness_bound
        ).result()
        if read_result.state != "value":
            raise KeyError(key)
        return self._value_decoder(key, read_result.value)

    def __contains__(self, key: K) -> bool:
        """Checks if a given key is present."""
        if self._kvstore is None:
            raise KeyError("required index kind not available")
        return self._key_encoder(key) in self._kvstore

    def get(self, key: K, batch: ts.Batch | None = None) -> ts.Future:
        """Reads a given key, returning a Future."""
        if self._kvstore is None:
            raise KeyError("required index kind not available")
        promise, future = ts.Promise.new()

        def done_callback(future):
            try:
                read_result = future.result()
                if read_result.state != "value":
                    promise.set_result(None)
                promise.set_result(self._value_decoder(key, read_result.value))
            except Exception as e:
                promise.set_exception(e)

        self._kvstore.read(
            self._key_encoder(key), staleness_bound=self.staleness_bound
        ).add_done_callback(done_callback)
        return future

    async def get_async(self, key: K, batch: ts.Batch | None = None) -> V:
        """Asynchronously reads a given key."""
        if self._kvstore is None:
            raise KeyError("required index kind not available")
        read_result = await self._kvstore.read(
            self._key_encoder(key), staleness_bound=self.staleness_bound, batch=batch
        )
        if read_result.state != "value":
            raise KeyError(key)
        return self._value_decoder(key, read_result.value)


AnnotationType = write_annotations.AnnotationType
"""Specifies the annotation type."""


def _decode_point_annotation(geom, rank: int, props, segments, id: str):
    return viewer_state.PointAnnotation(
        point=geom, props=props, segments=segments, id=id
    )


def _decode_line_annotation(geom, rank: int, props, segments, id: str):
    return viewer_state.LineAnnotation(
        point_a=geom[:rank], point_b=geom[rank:], props=props, segments=segments, id=id
    )


def _decode_axis_aligned_bounding_box_annotation(
    geom, rank: int, props, segments, id: str
):
    return viewer_state.AxisAlignedBoundingBoxAnnotation(
        point_a=geom[:rank], point_b=geom[rank:], props=props, segments=segments, id=id
    )


def _decode_ellipsoid_annotation(geom, rank: int, props, segments, id: str):
    return viewer_state.EllipsoidAnnotation(
        center=geom[:rank], radii=geom[rank:], props=props, segments=segments, id=id
    )


_ANNOTATION_TYPE_CONSTRUCTORS = {
    "point": _decode_point_annotation,
    "line": _decode_line_annotation,
    "axis_aligned_bounding_box": _decode_axis_aligned_bounding_box_annotation,
    "ellipsoid": _decode_ellipsoid_annotation,
}


def _point_check_bounds(annotation, lower_bound, upper_bound):
    point = annotation.point
    return np.all(point >= lower_bound) and np.all(point <= upper_bound)


def _bbox_check_bounds(min_pt, max_pt, lower_bound, upper_bound):
    return np.all(min_pt <= upper_bound) and np.all(max_pt <= lower_bound)


def _line_check_bounds(annotation, lower_bound, upper_bound):
    # For now, just perform a bounding box check for simplicity.
    # TODO(jbms): perform a more precise check
    point_a = annotation.point_a
    point_b = annotation.point_b
    min_pt = np.minimum(point_a, point_b)
    max_pt = np.maximum(point_a, point_b)
    return _bbox_check_bounds(min_pt, max_pt, lower_bound, upper_bound)


def _axis_aligned_bounding_box_check_bounds(annotation, lower_bound, upper_bound):
    point_a = annotation.point_a
    point_b = annotation.point_b
    min_pt = np.minimum(point_a, point_b)
    max_pt = np.maximum(point_a, point_b)
    return _bbox_check_bounds(min_pt, max_pt, lower_bound, upper_bound)


def _ellipsoid_check_bounds(annotation, lower_bound, upper_bound):
    center = annotation.point_a
    radii = annotation.radii
    return _bbox_check_bounds(center - radii, center + radii, lower_bound, upper_bound)


_ANNOTATION_TYPE_CHECK_BOUNDS = {
    "point": _point_check_bounds,
    "line": _line_check_bounds,
    "axis_aligned_bounding_box": _axis_aligned_bounding_box_check_bounds,
    "ellipsoid": _ellipsoid_check_bounds,
}


class AnnotationReader:
    """Provides read access to a Neuroglancer Precomputed annotation dataset."""

    _context: ts.Context

    coordinate_space: coordinate_space.CoordinateSpace
    """Coordinate space of annotations.

    Group:
      Accessors
    """

    staleness_bound: float
    """Staleness bound to use when reading.

    Group:
      Accessors
    """

    annotation_type: AnnotationType
    """Type of annotations.

    Group:
      Accessors
    """

    metadata: dict[str, typing.Any]
    """Raw Precomputed Annotation JSON metadata.

    Group:
      Accessors
    """

    properties: list[viewer_state.AnnotationPropertySpec]
    """Per-annotation properties included in the dataset.

    Group:
      Accessors
    """

    by_id: AnnotationMap[int, viewer_state.Annotation]
    """Annotations indexed by their 64-bit annotation id.

    If the by_id index is not present in `.metadata`, lookups raise errors.

    Group:
      I/O
    """

    relationships: dict[str, AnnotationMap[int, list[viewer_state.Annotation]]]
    """Related segment maps for each relationship.

    Group:
      I/O
    """

    lower_bound: tuple[float, ...]
    """Lower bound of all annotations within `.coordinate_space`.

    Group:
      Accessors
    """

    upper_bound: tuple[float, ...]
    """Upper bound of all annotations within `.coordinate_space`.

    Group:
      Accessors
    """

    def __init__(
        self,
        base_spec: typing.Any | ts.KvStore.Spec,
        context: ts.Context | None = None,
        staleness_bound: float | typing.Literal["open"] = "open",
    ):
        """Constructs an annotation reader.

        Args:
          base_spec: URL or TensorStore KvStore spec for the precomputed
              annotation directory.
          context: Optional TensorStore context to use.

          staleness_bound: Staleness bound for caching sharding metadata.  If not
              specified, defaults to the time at which the `AnnotationReader` is
              constructed.

        Group:
          Constructors
        """
        if context is None:
            context = ts.Context(
                {"cache_pool": {"total_bytes_limit": 1024 * 1024 * 20}}
            )
        self._context = context
        base_spec = ts.KvStore.Spec(base_spec)
        if base_spec.path and not base_spec.path.endswith("/"):
            base_spec.path += "/"
        self.base_spec = base_spec
        self.metadata = json.loads(
            ts.KvStore.open(self.base_spec, context=self._context).result()["info"]
        )
        self.lower_bound = tuple(self.metadata["lower_bound"])
        self.upper_bound = tuple(self.metadata["upper_bound"])
        if staleness_bound == "open":
            staleness_bound = time.time()
        self.staleness_bound = staleness_bound
        self.by_id = self._get_child_uint64_map(
            self.metadata["by_id"], self._decode_single_annotation
        )

        self.coordinate_space = coordinate_space.CoordinateSpace(
            self.metadata["dimensions"]
        )
        self.annotation_type = self.metadata["annotation_type"].lower()

        self.relationships = {
            relationship_metadata["id"]: self._get_child_uint64_map(
                relationship_metadata, self._decode_multiple_annotations
            )
            for relationship_metadata in self.metadata.get("relationships", [])
        }
        self.properties = [
            viewer_state.AnnotationPropertySpec(prop)
            for prop in self.metadata.get("properties", [])
        ]
        self._dtype = write_annotations._get_dtype_for_geometry(
            self.annotation_type, self.coordinate_space.rank
        ) + write_annotations._get_dtype_for_properties(self.properties)
        self.spatial = [
            self._get_child_spatial_map(spatial_metadata)
            for spatial_metadata in self.metadata.get("spatial", [])
        ]

    def _decode_single_annotation(
        self, annotation_id: int, encoded: bytes
    ) -> viewer_state.Annotation:
        decoded = np.frombuffer(encoded, dtype=self._dtype, count=1)[0]
        geom = decoded["geometry"]
        props = [decoded[f"property{i}"] for i in range(len(self.properties))]
        offset = decoded.nbytes
        segments = []
        for i in range(len(self.relationships)):
            count = np.frombuffer(encoded, dtype="<u4", count=1, offset=offset)[0]
            offset += 4
            segments.append(
                np.frombuffer(encoded, dtype="<u8", count=count, offset=offset)
            )
            offset += 8 * count
        return _ANNOTATION_TYPE_CONSTRUCTORS[self.annotation_type](
            geom, self.coordinate_space.rank, props, segments, str(annotation_id)
        )

    def _decode_multiple_annotations(
        self, unused_key, encoded: bytes
    ) -> list[viewer_state.Annotation]:
        count = np.frombuffer(encoded, dtype="<u8", count=1)[0]
        offset = 8
        decoded = np.frombuffer(encoded, dtype=self._dtype, count=count, offset=offset)
        offset += decoded.nbytes
        ids = np.frombuffer(encoded, dtype="<u8", count=count, offset=offset)
        offset += ids.nbytes
        if offset != len(encoded):
            raise ValueError(
                f"Expected encoded size to be {offset} bytes but actual size is {len(encoded)}"
            )
        constructor = _ANNOTATION_TYPE_CONSTRUCTORS[self.annotation_type]
        rank = self.coordinate_space.rank
        num_properties = len(self.properties)
        return [
            constructor(
                decoded[annotation_i]["geometry"],
                rank,
                [decoded[annotation_i][f"property{i}"] for i in range(num_properties)],
                None,
                str(ids[annotation_i]),
            )
            for annotation_i in range(count)
        ]

    def _get_child_kvstore(self, metadata: typing.Any) -> ts.KvStore | None:
        if metadata is None:
            return None
        base_spec = self.base_spec.copy()
        base_spec.path += metadata["key"] + "/"
        if "sharding" in metadata:
            return ts.KvStore.open(
                {
                    "driver": "neuroglancer_uint64_sharded",
                    "base": base_spec,
                    "metadata": metadata["sharding"],
                },
                context=self._context,
            ).result()
        return ts.KvStore.open(base_spec, context=self._context).result()

    def _get_child_uint64_map(
        self, metadata: typing.Any, value_decoder: typing.Callable[[K, bytes], V]
    ):
        return AnnotationMap(
            kvstore=self._get_child_kvstore(metadata),
            metadata=metadata,
            key_encoder=_get_uint64_key_encoder(metadata),
            value_decoder=value_decoder,
            staleness_bound=self.staleness_bound,
        )

    def _get_child_spatial_map(self, metadata):
        return AnnotationMap(
            kvstore=self._get_child_kvstore(metadata),
            metadata=metadata,
            key_encoder=_get_spatial_key_encoder(metadata),
            value_decoder=self._decode_multiple_annotations,
            staleness_bound=self.staleness_bound,
        )

    def get_within_spatial_bounds(
        self,
        *,
        lower_bound: typing.Sequence[float] | None = None,
        upper_bound: typing.Sequence[float] | None = None,
        min_spatial_index_level=0,
        limit: int | None = None,
        max_parallelism: int = 128,
    ) -> typing.Iterator[viewer_state.Annotation]:
        """Returns an iterator over the annotations within the specified bounds.

        Args:
          lower_bound: Lower bound within `.coordinate_space`.
              If not specified, defaults to `.lower_bound`.
          upper_bound: Upper bound within `.coordinate_space`.
              If not specified, defaults to `.upper_bound`.
          min_spatial_index_level: Minimum spatial index level to use.
          limit: Maximum number of iterations to return.

        Group:
          I/O
        """
        if lower_bound is None:
            lower_bound = self.lower_bound
        else:
            lower_bound = np.maximum(self.lower_bound, lower_bound)  # type: ignore[assignment]

        if upper_bound is None:
            upper_bound = self.upper_bound
        else:
            upper_bound = np.minimum(self.upper_bound, upper_bound)  # type: ignore[assignment]

        if np.any(upper_bound < lower_bound):  # type: ignore[operator]
            return

        check_bounds = _ANNOTATION_TYPE_CHECK_BOUNDS[self.annotation_type]

        count = 0
        outstanding_requests: collections.deque[
            tuple[
                ts.Future, np.typing.NDArray[np.float64], np.typing.NDArray[np.float64]
            ]
        ] = collections.deque()

        def handle_outstanding_request():
            nonlocal count

            future, req_lower_bound, req_upper_bound = outstanding_requests.pop()
            annotations = future.result()
            if annotations is None:
                return
            if np.all(req_lower_bound >= lower_bound) and np.all(
                req_upper_bound <= upper_bound
            ):
                annotation_iter = iter(annotations)
            else:
                annotation_iter = (
                    annotation
                    for annotation in annotations
                    if check_bounds(annotation, lower_bound, upper_bound)
                )
            for annotation in annotation_iter:
                if limit is not None and count >= limit:
                    break
                yield annotation
                count += 1

        def add_request(spatial_index, coords, req_lower_bound, req_upper_bound):
            while len(outstanding_requests) >= max_parallelism:
                yield from handle_outstanding_request()

            if limit is not None and count >= limit:
                return
            outstanding_requests.appendleft(
                (spatial_index.get(coords), req_lower_bound, req_upper_bound)
            )

        for level in range(len(self.spatial) - 1, min_spatial_index_level - 1, -1):
            if limit is not None and count >= limit:
                return
            spatial_index = self.spatial[level]
            metadata = spatial_index.metadata
            chunk_size = metadata["chunk_size"]
            grid_shape = metadata["grid_shape"]
            min_chunk = np.asarray(
                (np.array(lower_bound) - self.lower_bound) // chunk_size, dtype=np.int64
            )
            max_chunk = (
                np.asarray(
                    (np.array(upper_bound) - self.lower_bound) // chunk_size,
                    dtype=np.int64,
                )
                + 1
            )
            min_chunk = np.maximum(0, min_chunk)
            max_chunk = np.minimum(grid_shape, max_chunk)
            for coords in np.ndindex(*(max_chunk - min_chunk)):
                if limit is not None and count >= limit:
                    return
                coords = coords + min_chunk  # type: ignore[assignment]
                req_lower_bound = coords * chunk_size + self.lower_bound
                req_upper_bound = np.minimum(
                    self.upper_bound, req_lower_bound + chunk_size
                )
                yield from add_request(
                    spatial_index, coords, req_lower_bound, req_upper_bound
                )

        while outstanding_requests:
            yield from handle_outstanding_request()



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import concurrent.futures
import threading

from .futures import future_then_immediate


class CredentialsManager:
    def __init__(self):
        self._providers = dict()

    def register(self, key, credentials_provider_getter):
        self._providers[key] = credentials_provider_getter

    def get(self, key, parameters):
        return self._providers[key](parameters)


class CredentialsProvider:
    next_generation = 0
    next_generation_lock = threading.Lock()

    def __init__(self):
        self.credentials = None
        self.future = None
        self._lock = threading.Lock()

    def get(self, invalid_generation=None):
        with self._lock:
            if self.future is not None and (
                self.credentials is None
                or invalid_generation != self.credentials["generation"]
            ):
                return self.future
            self.credentials = None
            self.future = future = concurrent.futures.Future()

        def attach_generation_and_save_credentials(credentials):
            with self._lock:
                with CredentialsProvider.next_generation_lock:
                    CredentialsProvider.next_generation += 1
                    next_generation = CredentialsProvider.next_generation
                credentials_with_generation = dict(
                    credentials=credentials, generation=next_generation
                )
                self.credentials = credentials_with_generation
                return credentials_with_generation

        future_then_immediate(
            self.get_new(), attach_generation_and_save_credentials, future
        )
        return future

    def get_new(self):
        raise NotImplementedError



"""
MGH file format for affine transformations.

https://surfer.nmr.mgh.harvard.edu/fswiki/FsTutorial/LtaFormat

A bit must be said on the FreeSurfer space conventions:
- Any 3D volume is associated with a "physical" coordinate system.
  This system only depends on the voxel size and volume shape.
  This coordinate system is millimetric and ensures that the origin is
  in the center of the field-of-view.
- There is also a more generic "RAS" system. MGH images store
  an additional transform in an xform (= columns of a 3x4  matrix).
  This transform maps from physical to RAS space. When on-disk
  dimensions are not ordered as RAS, the xform includes shifts and
  permutations, to ensure that the output coordinates follow the RAS
  orientation.
- A NIfTI-like orientation matrix (named vox2ras in the FS nomenclature)
  can be recovered by composing the voxel-to-physical and the
  physical-to-ras matrices.
- Note that before registration (when images are mis-registered), their
  RAS might differ. Registration typically finds a ras-to-ras
  transform that brings the two dissociate spaces in alignment.
- The LTA file stores such a world-to-world transform, along with the
  orientation matrices of the original source and destination images.
  The voxel-to-physical and physical-to-ras components of these
  orientation matrices are stored as well.
- However, the LTA can also store transforms between different kinds of
  spaces (voxel-to-voxel, physical-to-physical, ras-to-ras or even
  rsa-to-rsa).
"""
# stdlib
import os
from dataclasses import dataclass
from enum import IntEnum
from typing import Iterator, Literal, TextIO, Union

# externals
import numpy as np

# internals
from .fsutils import (
    affine_matmul,
    affine_to_fs,
    fs_to_affine,
    read_key,
    read_values,
    write_key,
    write_values,
)


class Constants(IntEnum):
    """Defines used in FreeSurfer's C code."""

    # Affine transformation types
    LINEAR_VOX_TO_VOX = 0
    LINEAR_VOXEL_TO_VOXEL = LINEAR_VOX_TO_VOX
    LINEAR_RAS_TO_RAS = 1
    LINEAR_PHYSVOX_TO_PHYSVOX = 2
    LINEAR_CORONAL_RAS_TO_CORONAL_RAS = 21
    LINEAR_COR_TO_COR = LINEAR_CORONAL_RAS_TO_CORONAL_RAS
    # Transformation file types
    TRANSFORM_ARRAY_TYPE = 10
    MORPH_3D_TYPE = 11
    MNI_TRANSFORM_TYPE = 12
    MATLAB_ASCII_TYPE = 13


# Value format (= sequence of types) associated to each key
lta_keys = {
    'type': int,
    'nxforms': int,
    'mean': (float,) * 3,
    'sigma': float,
    'valid': int,
    'filename': str,
    'volume': (int,) * 3,
    'voxelsize': (float,) * 3,
    'xras': (float,) * 3,
    'yras': (float,) * 3,
    'zras': (float,) * 3,
    'cras': (float,) * 3,
}


def nested_update(old_dict: dict, new_dict: dict) -> None:
    """
    Update a dictionary in place.

    Nested dictionaries are updated too instead of being replaced.

    Parameters
    ----------
    old_dict : dict
        Dictionary updated in-place
    new_dict : dict
        New values

    """
    for key, value in new_dict.items():
        if (key in old_dict and isinstance(old_dict[key], dict) and
                isinstance(value, dict)):
            nested_update(old_dict[key], value)
        else:
            old_dict[key] = value


@dataclass
class LTAStruct:
    """Structure encoding an LTA file.

    This representation mimics the representation on dist and is not
    exposed to the user.
    """

    type: int = Constants.LINEAR_VOX_TO_VOX     # Affine type
    nxforms: int = 1                            # Number of affines stored
    mean: tuple = (0., 0., 0.)                  # ?
    sigma: float = 1.                           # ?
    affine: np.ndarray = None                   # Affine(s)

    @dataclass
    class VolumeInfo:
        valid: bool = 1                         #
        filename: str = None                    # Filename of the volume
        volume: tuple = None                    # 3D shape
        voxelsize: tuple = None                 # Voxel size
        xras: tuple = None                      # Columns of the xform
        yras: tuple = None                      # "
        zras: tuple = None                      # "
        cras: tuple = None                      # "

    src: VolumeInfo = None                      # Source volume
    dst: VolumeInfo = None                      # Destination volume

    def __str__(self) -> str:
        s = '\n'.join(list(self.to_lines()))
        s = 'LTAStruct\n' + '---------\n' + s
        return s

    __repr__ = __str__

    @classmethod
    def from_filename(cls, fname: str) -> "LTAStruct":
        """Build from path to LTA file."""
        with open(fname, 'r') as f:
            return cls.from_lines(f)

    @classmethod
    def from_lines(cls, f: list[str]) -> "LTAStruct":
        """Build from an iterator over lines."""
        lta = LTAStruct()
        section = 'header'
        affine = []
        affine_shape = (1, 4, 4)
        for line in f:
            if hasattr(line, 'decode'):
                line = line.decode()
            line = line.split('\r\n')[0]  # remove eol (windows)
            line = line.split('\n')[0]    # remove eol (unix)
            line = line.split('#')[0]     # remove hanging comments
            line = line.rstrip()          # remove trailing whitespaces
            if not line:
                continue
            if line.startswith('src volume info'):
                section = 'src0'
                continue
            elif line.startswith('dst volume info'):
                section = 'dst0'
                continue
            elif section == 'header':
                if '=' in line:
                    try:
                        key, value = read_key(line, lta_keys)
                        setattr(lta, key, value)
                    finally:
                        continue
                # else we should be in the affine section
                try:
                    affine_shape = read_values(line, [int] * 3)
                    if not affine_shape:
                        affine_shape = (1, 4, 4)
                    section = 'affine'
                finally:
                    continue
            elif section == 'affine':
                try:
                    row = read_values(line, [float] * affine_shape[-1])
                    if not affine:
                        affine.append([])
                    if len(affine[-1]) == affine_shape[-2]:
                        affine.append([])
                    affine[-1].append(list(row))
                finally:
                    continue
            else:
                if section.endswith('0'):
                    section = section[:-1]
                    setattr(lta, section, cls.VolumeInfo())
                vol = section
                try:
                    key, value = read_key(line, lta_keys)
                    setattr(getattr(lta, vol), key, value)
                finally:
                    continue
        if affine is not None:
            lta.affine = np.asarray(affine).reshape(affine_shape)
        return lta

    def to_lines(self) -> Iterator[str]:
        """Return an iterator over lines."""
        # header
        attributes = ('type', 'nxforms', 'mean', 'sigma')
        for attr in attributes:
            val = getattr(self, attr)
            if val is not None:
                row = write_key(attr, val)
                if attr == 'type':
                    row = row + '  # ' + Constants(val).name
                yield row
        # affine
        if self.affine is not None:
            yield write_values(self.affine.shape)
            affine = self.affine.reshape(-1, self.affine.shape[-1])
            for row in affine:
                yield write_values(row)
        # volume info
        volumes = ('src', 'dst')
        attributes = ('valid', 'filename', 'volume', 'voxelsize',
                      'xras', 'yras', 'zras', 'cras')
        for volume in volumes:
            block = getattr(self, volume)
            if block is None:
                continue
            yield f'{volume} volume info'
            for attr in attributes:
                val = getattr(block, attr)
                if val is not None:
                    yield write_key(attr, val)
        return

    def to_file_like(self, f: TextIO) -> None:
        """Write to file-like object."""
        for line in self.to_lines():
            f.write(line + os.linesep)
        return

    def to_filename(self, fname: str) -> None:
        """Write to file."""
        with open(fname, 'wt') as f:
            return self.to_file_like(f)

    def to(self, thing: str | TextIO) -> None:
        """Write to something."""
        if isinstance(thing, str):
            return self.to_filename(thing)
        else:
            return self.to_file_like(thing)


class LinearTransformArray:
    """MGH format for affine transformations."""

    @classmethod
    def possible_extensions(cls) -> tuple[str]:
        return ('.lta',)

    def __init__(
        self,
        file_like: str | TextIO | None = None,
        mode: str = 'r'
    ) -> None:
        """

        Parameters
        ----------
        file_like : str of file object
            File to map
        mode : {'r', 'r+'}, default='r'
            Read in read-only ('r') or read-and-write ('r+') mode.
            Modifying the file in-place is only possible in 'r+' mode.
        """
        self.filename = None
        self.mode = mode
        if file_like is None:
            self._struct = LTAStruct()
        elif isinstance(file_like, LTAStruct):
            self._struct = file_like
        else:
            self.filename = file_like
            if 'r' in mode:
                if isinstance(file_like, str):
                    self._struct = LTAStruct.from_filename(file_like)
                else:
                    self._struct = LTAStruct.from_lines(file_like)
            else:
                self._struct = LTAStruct()

    @property
    def shape(self) -> list[int]:
        if self._struct.affine is not None:
            return tuple(self._struct.affine.shape)
        else:
            return tuple()

    def matrix(
        self,
        source: Literal['voxel', 'physical', 'ras'] = 'ras',
        dest: Literal['voxel', 'physical', 'ras'] = 'ras',
        dtype: np.dtype | None = None
    ) -> np.ndarray:
        """Return <source>2<dest> matrix.

        Parameters
        ----------
        source : {'voxel', 'physical', 'ras'}, default='ras'
        dest : {'voxel', 'physical', 'ras'}, default='ras'
        dtype : np.dtype, optional

        Returns
        -------
        affine : (D+1, D+1) array[dtype]

        """
        backend = dict(dtype=dtype)
        affine = self.raw_matrix(**backend)
        if affine is None:
            return None

        # we may need to convert from a weird space to RAS space
        if self.type()[0] != source or self.type()[1] != dest:
            src, _ = self.source_space(source, self.type()[0], **backend)
            dst, _ = self.destination_space(self.type()[1], dest, **backend)
            if src is not None and dst is not None:
                affine = affine_matmul(dst, affine_matmul(affine, src))
        return affine

    def raw_matrix(self, dtype: np.dtype | None = None) -> np.ndarray:
        """Return raw matrix."""
        if self._struct.affine is None:
            return None
        return self._struct.affine.astype(dtype)

    def source_space(
        self,
        source: Literal['voxel', 'physical', 'ras'] = 'voxel',
        dest: Literal['voxel', 'physical', 'ras'] = 'ras',
        dtype: np.dtype | None = None
    ) -> tuple[np.ndarray, list[int]]:
        """Return the space (affine + shape) of the source image.

        Parameters
        ----------
        source : {'voxel', 'physical', 'ras'}, default='voxel'
            Source space of the affine
        dest : {'voxel', 'physical', 'ras'}, default='ras'
            Destination space of the affine
        dtype : torch.dtype, optional

        Returns
        -------
        affine : (4, 4) array
            A voxel to world affine matrix
        shape : (3,) tuple[int]
            The spatial shape of the image

        """
        if self._struct.src is not None:
            affine = fs_to_affine(
                self._struct.src.volume,
                self._struct.src.voxelsize,
                self._struct.src.xras,
                self._struct.src.yras,
                self._struct.src.zras,
                self._struct.src.cras,
                source=source, dest=dest)
            shape = tuple(self._struct.src.volume)
            affine = affine.astype(dtype)
            return affine, shape
        return None, None

    def destination_space(
        self,
        source: Literal['voxel', 'physical', 'ras'] = 'voxel',
        dest: Literal['voxel', 'physical', 'ras'] = 'ras',
        dtype: np.dtype | None = None
    ) -> tuple[np.ndarray, list[int]]:
        """Return the space (affine + shape) of the destination image.

        Parameters
        ----------
        source : {'voxel', 'physical', 'ras'}, default='voxel'
            Source space of the affine
        dest : {'voxel', 'physical', 'ras'}, default='ras'
            Destination space of the affine
        dtype : torch.dtype, optional

        Returns
        -------
        affine : (4, 4) array
            A voxel to world affine matrix
        shape : (3,) tuple[int]
            The spatial shape of the image

        """
        if self._struct.dst is not None:
            affine = fs_to_affine(
                self._struct.dst.volume,
                self._struct.dst.voxelsize,
                self._struct.dst.xras,
                self._struct.dst.yras,
                self._struct.dst.zras,
                self._struct.dst.cras,
                source=source, dest=dest)
            shape = tuple(self._struct.dst.volume)
            affine = affine.astype(dtype)
            return affine, shape
        return None, None

    def set_source_space(
        self,
        affine: np.ndarray,
        shape: list[int],
        source: Literal['voxel', 'physical', 'ras'] = 'voxel',
        dest: Literal['voxel', 'physical', 'ras'] = 'ras',
    ) -> "LinearTransformArray":
        """Set the source space of the transform.

        Parameters
        ----------
        affine : (4, 4) array
            Affine matrix
        shape : sequence of int
            Volume shape
        source : {'voxel', 'physical', 'ras'}, default='voxel'
            Source space of the affine
        dest : {'voxel', 'physical', 'ras'}, default='ras'
            Destination space of the affine

        Returns
        -------
        self

        """
        if not self._struct.src:
            self._struct.src = self._struct.VolumeInfo()
        if not self._struct.dst:
            self._struct.dst = self._struct.VolumeInfo()
        vx, x, y, z, c = affine_to_fs(affine, shape, source, dest)
        self._struct.src.volume = tuple(shape)
        self._struct.src.voxelsize = vx
        self._struct.src.xras = x
        self._struct.src.yras = y
        self._struct.src.zras = z
        self._struct.src.cras = c
        return self

    def set_destination_space(
        self,
        affine: np.ndarray,
        shape: list[int],
        source: Literal['voxel', 'physical', 'ras'] = 'voxel',
        dest: Literal['voxel', 'physical', 'ras'] = 'ras',
    ) -> "LinearTransformArray":
        """Set the destination space of the transform.

        Parameters
        ----------
        affine : (4, 4) array
            Affine matrix
        shape : sequence of int
            Volume shape
        source : {'voxel', 'physical', 'ras'}, default='voxel'
            Source space of the affine
        dest : {'voxel', 'physical', 'ras'}, default='ras'
            Destination space of the affine

        Returns
        -------
        self

        """
        vx, x, y, z, c = affine_to_fs(affine, shape, source, dest)
        self._struct.dst.volume = tuple(shape)
        self._struct.dst.voxelsize = vx
        self._struct.dst.xras = x
        self._struct.dst.yras = y
        self._struct.dst.zras = z
        self._struct.dst.cras = c
        return self

    def type(self) -> Constants:
        if self._struct.type == Constants.LINEAR_VOX_TO_VOX:
            return 'voxel', 'voxel'
        elif self._struct.type == Constants.LINEAR_RAS_TO_RAS:
            return 'ras', 'ras'
        elif self._struct.type == Constants.LINEAR_COR_TO_COR:
            return 'rsa', 'rsa'
        elif self._struct.type == Constants.LINEAR_PHYSVOX_TO_PHYSVOX:
            return 'physical', 'physical'
        else:
            raise TypeError(f'Don\'t know what to do with {self._struct.type}')

    def metadata(self, keys: list[str] | None = None) -> dict:
        """Read additional metadata.

        Parameters
        ----------
        keys : sequence of str, optional
            Keys should be in {'type', 'mean', 'sigma', 'src',
                'src',
                'src.valid', 'src.filename', 'src.volume', 'src.voxelsize',
                'src.xras', 'src.yras', 'src.zras', 'src.cras',
                'dst',
                'dst.valid', 'dst.filename', 'dst.volume', 'dst.voxelsize',
                'dst.xras', 'dst.yras', 'dst.zras', 'dst.cras'
            }

        Returns
        -------
        dict
            type : {'ras', 'rsa', 'physical', 'voxel'}
                Transformation type
            mean : (3,) sequence of float
                Mean image intensity
            sigma : float
                Standard deviation of the image intensity
            src : dict
                Source volume information
            dst : dict
                Destination volume information
            src|dst.valid : bool
                Is the volume valid
            src|dst.filename : str
                Filename of the source volume
            src|dst.volume : tuple[int]
                3D shape
            src|dst.voxelsize : tuple[float]
                Voxel size
            src|dst.xras : tuple[float]
            src|dst.yras : tuple[float]
            src|dst.zras : tuple[float]
            src|dst.cras : tuple[float]

        """
        known_keys = ('type', 'mean', 'sigma')
        known_superkeys = ('src', 'dst')
        known_subkeys = ('valid', 'filename', 'volume', 'voxelsize',
                         'xras', 'yras', 'zras', 'cras')
        all_keys = known_keys + tuple(sup + '.' + sub
                                      for sub in known_subkeys
                                      for sup in known_superkeys)
        keys = keys or all_keys
        meta = dict()
        for key in keys:
            if key in known_keys:
                meta[key] = getattr(self._struct, key)
            else:
                sup, *sub = key.split('.')
                if sup in known_superkeys and sub and sub[0] in known_subkeys:
                    sub = sub[0]
                    meta[key] = getattr(getattr(self._struct, sup), sub)
                else:
                    meta[key] = None
        return meta

    def set_fdata(self, affine: np.ndarray) -> "LinearTransformArray":
        affine = np.asarray(affine)
        backend = dict(dtype=affine.dtype)
        if affine.shape[-2:] != (4, 4):
            raise ValueError('Expected a batch of 4x4 matrix')

        # we may need to convert from RAS space to a weird space
        afftype = self.type()[0]
        if afftype != 'ras':
            src, _ = self.source_space(afftype, 'ras', **backend)
            dst, _ = self.destination_space('ras', afftype, **backend)
            if src is not None and dst is not None:
                affine = affine_matmul(dst, affine_matmul(affine, src))

        affine = np.asarray(affine).reshape([-1, 4, 4])
        self._struct.affine = affine
        self._struct.nxform = affine.shape[0]
        return self

    def set_data(self, affine: np.ndarray) -> "LinearTransformArray":
        affine = np.asarray(affine)
        if affine.shape != (4, 4):
            raise ValueError('Expected a 4x4 matrix')
        affine = affine.reshape([-1, 4, 4])
        self._struct.affine = affine
        self._struct.nxform = affine.shape[0]
        return self

    @classmethod
    def _set_metadata(cls, struct: LTAStruct, **meta) -> None:
        """Set LTA fields (in-place) from a dictionary.

        Parameters
        ----------
        struct : LTAStruct
        meta : dict

        """
        if 'type' in meta and not isinstance(meta['type'], int):
            ltatype = meta['type']
            if isinstance(ltatype, (list, tuple)):
                srctype, dsttype = ltatype
            else:
                srctype = dsttype = ltatype
            if srctype == dsttype == 'ras':
                ltatype = Constants.LINEAR_RAS_TO_RAS
            elif srctype == dsttype == 'rsa':
                ltatype = Constants.LINEAR_COR_TO_COR
            elif srctype == dsttype == 'voxel':
                ltatype = Constants.LINEAR_VOX_TO_VOX
            elif srctype == dsttype == 'physical':
                ltatype = Constants.LINEAR_PHYSVOX_TO_PHYSVOX
            else:
                raise ValueError('Unsupported LTA type:', ltatype)
            meta['type'] = ltatype

        def tupleof(typ: type) -> callable:
            return (lambda x: tuple(typ(y) for y in x))

        known_keys = dict(type=int, mean=tupleof(float), sigma=float)
        known_superkeys = ('src', 'dst')
        known_subkeys = dict(valid=int, filename=str, volume=tupleof(int),
                             voxelsize=tupleof(float), xras=tupleof(float),
                             yras=tupleof(float), zras=tupleof(float),
                             cras=tupleof(float))
        for key, value in meta.items():
            if key in tuple(known_keys.keys()):
                conv = known_keys[key]
                setattr(struct, key, conv(value))
            elif key in known_superkeys and isinstance(value, dict):
                sup = getattr(struct, key)
                for subkey, subval in value.items():
                    if subkey in known_subkeys:
                        conv = known_subkeys[subkey]
                        setattr(sup, subkey, conv(subval))

    def set_metadata(
        self,
        dict_like: dict | None = None,
        **meta
    ) -> "LinearTransformArray":
        """Set additional metadata.

        Parameters
        ----------
        type : {'ras', 'rsa', 'physical', 'voxel'}
            Transformation type
        mean : (3,) sequence of float
            Mean image intensity
        sigma : float
            Standard deviation of the image intensity
        src, dst : dict
            Source/Destination volume information with keys:
            valid : bool
                Is the volume valid
            filename : str
                Filename of the source volume
            volume : tuple[int]
                3D shape
            voxelsize : tuple[float]
                Voxel size
            xras : tuple[float]
            yras : tuple[float]
            zras : tuple[float]
            cras : tuple[float]

        Returns
        -------
        self

        """
        dict_like = dict_like or {}
        dict_like.update(meta)
        self._set_metadata(self._struct, **dict_like)
        return self

    def save(
        self, file_like: str | TextIO | None = None, *args, **meta
    ) -> "LinearTransformArray":
        if '+' not in self.mode and 'w' not in self.mode:
            raise RuntimeError('Cannot write into read-only volume. '
                               'Re-map in mode "r+" to allow in-place '
                               'writing.')
        self.set_metadata(**meta)
        file_like = file_like or self.filename or self.file_like
        self._struct.to(file_like)
        return self

    @classmethod
    def save_new(
        cls,
        affine: Union[np.ndarray, "LinearTransformArray"],
        file_like: str | TextIO | None = None,
        like: Union["LinearTransformArray", str, TextIO] | None = None,
        *args,
        **meta
    ) -> "LinearTransformArray":
        if isinstance(affine, LinearTransformArray):
            if like is None:
                like = affine
            affine = affine.data()
        affine = np.asanyarray(affine)
        if like is not None and not isinstance(like, LinearTransformArray):
            like = LinearTransformArray(like)

        if affine.shape[-2:] != (4, 4):
            raise ValueError('Expected a batch of 4x4 matrix')
        affine = affine.reshape([-1, 4, 4])

        struct = LTAStruct()
        struct.affine = affine
        metadata = like.metadata() if like is not None else dict()
        nested_update(metadata, meta)
        cls._set_metadata(struct, **metadata)
        struct.to(file_like)



"""A `fsspec` File System for (remote) DANDI."""
# stdlib
import json
import logging
import os
import re
from os import PathLike
from typing import Iterator
from urllib.parse import unquote as url_unquote

# externals
import requests
from dandi.dandiapi import (
    BaseRemoteAsset,
    DandiAPIClient,
    DandiInstance,
    NotFoundError,
    RemoteAsset,
    RemoteDandiset,
)
from dandi.utils import get_instance
from fsspec import register_implementation
from fsspec.implementations.http import HTTPFileSystem
from fsspec.registry import known_implementations
from fsspec.spec import AbstractBufferedFile, AbstractFileSystem
from fsspec.utils import stringify_path, tokenize

LOG = logging.getLogger(__name__)


class RemoteDandiFileSystem(AbstractFileSystem):
    """
    A file system that browses through a remote dandiset.

    Examples
    --------
    Load and parse a remote file
    ```python
    from dandi.fs import RemoteDandiFileSystem
    import json
    fs = RemoteDandiFileSystem()
    with fs.open('dandi://dandi/000026/rawdata/sub-I38/ses-MRI/anat/'
                 'sub-I38_ses-MRI-echo-4_flip-4_VFA.json') as f:
        info = json.load(f)
    ```

    The 'dandi://' protocol is registered with fsspec, so the same
    result can be achived by
    ```python
    import fsspec
    import json
    with fsspec.open('dandi://dandi/000026/rawdata/sub-I38/ses-MRI/anat/'
                     'sub-I38_ses-MRI-echo-4_flip-4_VFA.json') as f:
        info = json.load(f)
    ```

    Browse a dataset
    ```python
    from dandi.fs import RemoteDandiFileSystem
    fs = RemoteDandiFileSystem('000026')
    fs.glob('**/anat/*.json')
    ```

    """

    def __init__(
        self,
        dandiset: str | RemoteDandiset | None = None,
        version: str | None = None,
        client: str | DandiInstance | DandiAPIClient | None = None,
        **http_kwargs
    ) -> None:
        """
        Initialise a remote DANDI file system.

        The root of a DANDI file system is a dandiset at a given version.
        The file system can be initialized from
            - a `RemoteDandiset` instance; or
            - the name of a dandiset [+ version]; and
                . a DandiAPIClient instance
                . a DandiInstance instance
                . the name of a known DANDI instance
                . the url of a DANDI server

        Parameters
        ----------
        dandiset : str or dandi.RemoteDandiset, optional
            An instantiated dandiset, or the identifier of a dandiset
            (e.g., `'000026'`). See `RemoteDandiset` for more info.
        version : str, optional
            The version of the dandiset to query (e.g., `'draft'`)
        client : str or dandi.DandiInstance or dandi.DandiAPIClient
            An instantiated dandi instance (or its identifier) or
            an instantiated dandi client (or its url).
            Default: `'dandi'`. See `DandiAPIClient` for more info.

        Other Parameters
        ----------------
        http_kwargs: key-value
            Any other parameters passed on to the HTTP file system
        """
        self._httpfs = HTTPFileSystem(**http_kwargs)
        super().__init__()
        if not isinstance(dandiset, RemoteDandiset):
            if isinstance(client, str):
                if not client.startswith('http'):
                    client = get_instance(client)
            if isinstance(client, DandiInstance):
                client = DandiAPIClient.for_dandi_instance(client)
            else:
                client = DandiAPIClient(client)
            if dandiset:
                dandiset = self.get_dandiset(client, dandiset, version)
        self._dandiset = dandiset
        self._client = None if dandiset else client

    # ------------------------------------------------------------------
    #   DANDI-specific helpers
    # ------------------------------------------------------------------

    @property
    def dandiset(self) -> RemoteDandiset:
        """Access dandiset."""
        return self._dandiset

    @dandiset.setter
    def dandiset(self, x: RemoteDandiset) -> None:
        """Assign dandiset."""
        if x:
            self._client = None
        elif self._dandiset:
            self._client = self._dandiset.client
        self._dandiset = x

    @property
    def client(self) -> DandiAPIClient:
        """Access dandi client."""
        return self.dandiset.client if self.dandiset else self._client

    @client.setter
    def client(self, x: DandiAPIClient) -> None:
        """Assign dandi client."""
        if self.dandiset:
            raise ValueError('Cannot assign a DANDI client to a FileSystem '
                             'that is already linked to a dandiset. '
                             'Unassign the dandiset first.')
        self._client = x

    @classmethod
    def for_url(cls, url: str) -> "RemoteDandiFileSystem":
        """
        Instantiate a FileSystem that interacts with the correct
        DANDI instance for a given url.
        """
        instance, dandiset, version, *_ = split_dandi_url(url)
        return cls(dandiset, version, instance)

    @classmethod
    def _auth_apply(
        cls, fn: callable, client: DandiAPIClient, auth: bool | None = None
    ) -> object:
        is_401 = lambda e: (  # noqa: E731
            e.response is not None
            and e.response.status_code == 401
            and auth is not False
        )
        level = logging.getLogger("dandi").getEffectiveLevel()
        logging.getLogger("dandi").setLevel(1000)
        try:
            if auth is not False:
                client.dandi_authenticate()
            return fn()
        except requests.HTTPError as e:
            if not is_401(e):
                raise e
            try:
                client.authenticate(os.environ.get("LINCBRAIN_API_KEY", ""))
                return fn()
            except requests.HTTPError as e:
                if not is_401(e):
                    raise e
                exc = e
            raise exc
        finally:
            logging.getLogger("dandi").setLevel(level)

    @classmethod
    def _get_dandiset(
        cls, client: DandiAPIClient, *args, **kwargs
    ) -> RemoteDandiset:
        auth = kwargs.pop("auth", None)
        return cls._auth_apply(
            lambda: client.get_dandiset(*args, **kwargs),
            client, auth
        )

    @classmethod
    def _get_asset(cls, client: DandiAPIClient, *a, **k) -> BaseRemoteAsset:
        auth = k.pop("auth", None)
        return cls._auth_apply(
            lambda: client.get_asset(*a, **k),
            client, auth
        )

    def get_dandiset(
        self, path: str, auth: bool | None = None,
        dandiset: RemoteDandiset | None = None,
    ) -> tuple[RemoteDandiset, str]:
        """
        If path is a relative path, return (self.dandiset, path)
        Else, the path is an absolute URL and we instantiate the correct
        remote dandiset and spit out the relative path.

        Returns: (dandiset, path) or (dandiset, asset)
        """
        dandiset = dandiset or self.dandiset
        if path.startswith(('http://', 'https://', 'dandi://', 'DANDI:')):
            instance, dandiset_id, version_id, path, asset_id \
                = split_dandi_url(path)
            api_url = get_instance(instance)
            if self.client.api_url == api_url.api:
                client = self.client
            else:
                client = DandiAPIClient.for_dandi_instance(instance)
                dandiset = None
            if not asset_id:
                args = (dandiset_id, version_id)
                if not dandiset or dandiset.identifier != dandiset_id:
                    dandiset = self._get_dandiset(client, *args, auth=auth)
                if not dandiset or dandiset.version_id != version_id:
                    dandiset = self._get_dandiset(client, *args, auth=auth)
            else:
                asset = self._get_asset(client, asset_id, auth=auth)
                return dandiset, asset
        elif not dandiset:
            raise ValueError('File system must be linked to a dandiset to '
                             'use relative paths.')
        return dandiset, path

    def _get_json(self, url: str) -> dict:
        with self._httpfs.open(url, "rt") as f:
            info = json.load(f)
        return info

    def _s3_url_from_asset(self, asset: RemoteAsset) -> str:
        info = self._get_json(asset.api_url)
        # info = requests.request(url=asset.api_url, method='get').json()
        url = ''
        if "neuroglancerUrl" in info:
            # LINC: use neuroglancer url but drop format protocol
            url = info["neuroglancerUrl"]
            url = url.split("://")
            url = "://".join(url[1:])
            if 'neuroglancer.lincbrain.org' not in url:
                raise NotFoundError(asset, url)
            return url
        for url in info['contentUrl']:
            if "s3.amazonaws.com" in url:
                return url
        raise NotFoundError(asset, url)

    def _maybe_to_s3(self, url: str) -> str:
        url = stringify_path(url)
        # FIXME: not very generic test
        is_s3 = (
            's3.amazonaws.com' in url or
            'neuroglancer.lincbrain.org' in url
        )
        if not is_s3:
            url = self.s3_url(url)
        return url

    def s3_url(self, path: str, **kwargs) -> str:
        """Get the the asset url on AWS S3."""
        dandiset, asset = self.get_dandiset(path, **kwargs)
        if not isinstance(path, RemoteAsset):
            try:
                asset = self._auth_apply(
                    lambda: dandiset.get_asset_by_path(asset),
                    dandiset.client, kwargs.get("auth", None)
                )

            except NotFoundError:
                path = asset.rstrip("/")
                path_prefix, path_suffix = path, ''
                while path_prefix:

                    *path_prefix, new_suffix = path_prefix.split('/')
                    if not path_suffix:
                        path_suffix = new_suffix
                    else:
                        path_suffix = path_suffix.split('/')
                        path_suffix = '/'.join([new_suffix, *path_suffix])
                    path_prefix = '/'.join(path_prefix)

                    try:
                        asset = self._auth_apply(
                            lambda: dandiset.get_asset_by_path(path_prefix),
                            dandiset.client, kwargs.get("auth", None)
                        )
                        url = self._s3_url_from_asset(asset).rstrip("/")
                        url += "/" + path_suffix
                        return url
                    except NotFoundError:
                        continue
                raise NotFoundError(path)

        return self._s3_url_from_asset(asset)

    # ------------------------------------------------------------------
    #   FileSystem API
    # ------------------------------------------------------------------

    def ls(  # noqa: D102
        self,
        path: str | PathLike,
        detail: bool = True,
        **kwargs
    ) -> list[str] | list[dict]:
        path = stringify_path(path).strip('/')
        assets = kwargs.pop('assets', None)
        if assets is None:
            dandiset = kwargs.pop('dandiset', None)
            if not dandiset:
                dandiset, path = self.get_dandiset(path)
            assets = dandiset.get_assets_with_path_prefix(path)

        entries = []
        full_dirs = set()

        def getdate(asset: RemoteAsset, field: str) -> str:
            return getattr(getattr(asset, field, None),
                           'isoformat', lambda: None)()

        assets, assets_in = [], assets
        for asset in assets_in:
            size = getattr(asset, 'size', None)
            created = getdate(asset, 'created')
            modified = getdate(asset, 'modified')
            identifier = getattr(asset, 'identifer', None)
            asset = getattr(asset, 'path', asset)
            # 1) is the input path exactly this asset?
            asset = asset[len(path):].strip('/')
            if not asset:
                entries.append({
                    'name': path,
                    'size': size,
                    'created': created,
                    'modified': modified,
                    'identifier': identifier,
                    'type': 'file',
                })
                continue
            # 2) look at the first level under `path`
            name = asset.split('/')[0]
            fullpath = path + '/' + name
            if '/' not in asset:
                # 3) this asset is a file directly under `path`
                entries.append({
                    'name': fullpath,
                    'size': size,
                    'created': created,
                    'modified': modified,
                    'identifier': identifier,
                    'type': 'file',
                })
                continue
            else:
                # 4) this asset is a file a few levels under `path`
                # -> we do not list the path but list the directory
                if fullpath not in full_dirs:
                    entries.append({
                        'name': fullpath,
                        'size': None,
                        'type': 'directory',
                    })
                    full_dirs.add(fullpath)
            assets.append(path + '/' + asset)

        if detail:
            return entries
        else:
            return [entry['name'] for entry in entries]

    def checksum(self, path: str, **kwargs) -> str:  # noqa: D102
        # we override fsspec's default implementation when path is a
        # directory (since in this case there is no created/modified date)
        dandiset = kwargs.pop('dandiset', None)
        if not dandiset:
            dandiset, path = self.get_dandiset(path)
        assets = dandiset.get_assets_with_path_prefix(path)
        return tokenize(assets)

    def glob(  # noqa: D102
        self,
        path: str,
        order: str | None = None,
        **kwargs
    ) -> Iterator[str]:
        # we override fsspec's default implementation (which uses find)
        # to leverage the more efficient `get_assets_by_glob` from dandi
        #
        # order : [-]{created, modified, path}
        #
        # TODO: implement fsspec `maxdepth` keyword
        dandiset = kwargs.pop('dandiset', None)
        if not dandiset:
            dandiset, path = self.get_dandiset(path)
        assets = dandiset.get_assets_by_glob(path, order)
        for asset in assets:
            yield asset.path

    def exists(self, path: str, **kwargs) -> bool:  # noqa: D102
        # we override fsspec's default implementation (which uses info)
        # to avoid calls to ls (which calls get_assets_by_path on the
        # *parent* and is therefore slower)
        dandiset = kwargs.pop('dandiset', None)
        if not dandiset:
            dandiset, path = self.get_dandiset(path)
        if isinstance(path, BaseRemoteAsset):
            return True
        # check if it is a file
        try:
            dandiset.get_asset_by_path(path)
            return True
        except NotFoundError:
            pass
        # check if it is a directory
        path = path.rstrip('/') + '/'
        assets = dandiset.get_assets_with_path_prefix(path)
        try:
            next(assets)
            return True
        except StopIteration:
            pass
        # it might be a path to something inside a zarr -- let's try to find it
        try:
            return self._httpfs.exists(self.s3_url(path, dandiset=dandiset))
        except NotFoundError:
            pass
        return False

    def open(  # noqa: D102
        self,
        path: str,
        *args,
        **kwargs
    ) -> AbstractBufferedFile:
        path = self._maybe_to_s3(path)
        return self._httpfs.open(path, *args, **kwargs)


def split_dandi_url(url: str) -> tuple[str, str, str, str, str]:
    """
    Split a valid dandi url into its subparts.
    Returns: (instance, dandiset_id, version_id, path, asset_id)
    where instance can be an instance_id or an URL.
    """
    instance = None
    server = None
    dandiset_id = None
    version = None
    path = ''
    asset_id = None
    if url.startswith('dandi://'):
        # dandi://<instance name>/<dandiset id>[@<version>][/<path>]
        ptrn = r'dandi://(?P<i>[^/]+)/(?P<d>\d+)(@(?P<v>[^/]+))?(?P<p>.*)'
        match = re.match(ptrn, url)
        if not match:
            raise SyntaxError('Wrong dandi url')
        instance = match.group('i')
        dandiset_id = match.group('d')
        version = match.group('v')
        path = match.group('p')
    elif url.startswith(('DANDI:', 'https://identifiers.org/DANDI:')):
        # DANDI:<dandiset id>[/<version id>]
        # https://identifiers.org/DANDI:<dandiset id>[/<version id>]
        ptrn = r'(https://identifiers.org/)?DANDI:(?P<d>\d+)(/(?P<v>[^/]+))?'
        match = re.match(ptrn, url)
        if not match:
            raise SyntaxError('Wrong dandi url')
        dandiset_id = match.group('d')
        version = match.group('v')
        instance = 'DANDI'
    else:
        ptrn = r'https://(?P<s>[^/]+)(/api)?(/#)?(?P<u>.*)'
        match = re.match(ptrn, url)
        if not match:
            raise SyntaxError('Wrong dandi url')
        server = match.group('s')
        url = match.group('u')
        if url.startswith('/dandisets/'):
            # https://<server>[/api]/dandisets/<dandiset id>
            #   . [/versions[/<version>]]
            #   . /versions/<version>/assets/<asset id>[/download]
            #   . /versions/<version>/assets/?path=<path>
            ptrn = r'/dandisets/(?P<d>\d+)(/versions/(?P<v>[^/]+))?(?P<u>.*)'
            match = re.match(ptrn, url)
            if not match:
                raise SyntaxError('Wrong dandi url')
            dandiset_id = match.group('d')
            version = match.group('v')
            url = match.group('u')
            ptrn = r'/assets/((\?path=(?P<p>[.*]+))|(?P<a>[^/]+))'
            match = re.match(ptrn, url)
            if match:
                path = match.group('p')
                asset_id = match.group('a')
        elif url.startswith('/dandiset/'):
            # https://<server>[/api]/[#/]dandiset/<dandiset id>
            #   [/<version>][/files[?location=<path>]]
            ptrn = r'(/(?P<v>[^/]+))?/files(\?location=(?P<p>.*))?'
            ptrn = r'/dandiset/(?P<d>\d+)' + ptrn
            match = re.match(ptrn, url)
            dandiset_id = match.group('d')
            version = match.group('v')
            path = match.group('p')
        elif url.startswith('/assets/'):
            # https://<server>[/api]/assets/<asset id>[/download]
            ptrn = r'/assets/(?P<a>[^/]+)'
            match = re.match(ptrn, url)
            if not match:
                raise SyntaxError('Wrong dandi url')
            asset_id = match.group('a')

    path = url_unquote(path)
    path = (path or '').strip('/')

    if instance is None:
        instance = 'https://' + server

    return instance, dandiset_id, version, path, asset_id


if "dandi" not in known_implementations:
    register_implementation("dandi", RemoteDandiFileSystem)
